"""
       █████  █████ ██████   █████           █████ █████   █████ ██████████ ███████████    █████████  ██████████
      ░░███  ░░███ ░░██████ ░░███           ░░███ ░░███   ░░███ ░░███░░░░░█░░███░░░░░███  ███░░░░░███░░███░░░░░█
       ░███   ░███  ░███░███ ░███   ██████   ░███  ░███    ░███  ░███  █ ░  ░███    ░███ ░███    ░░░  ░███  █ ░ 
       ░███   ░███  ░███░░███░███  ░░░░░███  ░███  ░███    ░███  ░██████    ░██████████  ░░█████████  ░██████   
       ░███   ░███  ░███ ░░██████   ███████  ░███  ░░███   ███   ░███░░█    ░███░░░░░███  ░░░░░░░░███ ░███░░█   
       ░███   ░███  ░███  ░░█████  ███░░███  ░███   ░░░█████░    ░███ ░   █ ░███    ░███  ███    ░███ ░███ ░   █
       ░░████████   █████  ░░█████░░████████ █████    ░░███      ██████████ █████   █████░░█████████  ██████████
        ░░░░░░░░   ░░░░░    ░░░░░  ░░░░░░░░ ░░░░░      ░░░      ░░░░░░░░░░ ░░░░░   ░░░░░  ░░░░░░░░░  ░░░░░░░░░░ 
                 A Collectionless AI Project (https://collectionless.ai)
                 Registration/Login: https://unaiverse.io
                 Code Repositories:  https://github.com/collectionlessai/
                 Main Developers:    Stefano Melacci (Project Leader), Christian Di Maio, Tommaso Guidi
"""
import math
import torch
from .psi import psi
import torch.nn.functional as F


class CNUs(torch.nn.Module):

    def __init__(self, q=1, d=2, m=3, u=4, delta=3,
                 gamma_alpha=0.1, tau_alpha=0.5, tau_mu=100, tau_eta=100,
                 upd_m="WTA", upd_k="ad_hoc_WTA",
                 beta_k=0.001,
                 psi_fn="identity",
                 scramble=False):
        """
        :param q: number of neurons
        :param d: size of each key
        :param m: number of keys/memory units
        :param u: size of each memory unit
        :param gamma_alpha: softmax temperature (key matching)
        :param tau_alpha: threshold on the attention score of the winning key, to eventually trigger scrambling
        :param tau_mu: number of steps below which a key is considered to be not-used enough
        :param tau_eta: number of steps after which a key is considered old
        :param delta: number of top attention responses to select (top-delta)
        :param upd_m: update memory strategy (None, 'WTA')
        :param upd_k: update key strategy (None, 'ad_hoc_WTA', 'grad_WTA')
        :param beta_k: learning rate for key-update purposes when upd_k is 'ad_hoc_WTA'
        :param psi_fn: function to project the neuron input onto the key space
        :param scramble: triggers the key/memory scrambling routine when upd_k is 'ad_hoc_WTA'
        """

        super(CNUs, self).__init__()
        assert upd_m in (None, 'WTA'), "Unknown value for upd_m, it must be None or 'WTA'"
        assert upd_k in (None, 'ad_hoc_WTA', 'grad_WTA'), "Unknown value for upd_k, it must be " \
                                                          "None, 'ad_hoc_WTA', or 'grad_WTA'"
        assert upd_m is None or (upd_m == 'WTA' and upd_k is not None), \
            "If upd_m is 'WTA', then upd_k must be ad_hoc_WTA or grad_WTA (it cannot be None)"
        self.q = q
        self.d = d
        self.m = m
        self.u = u
        self.gamma_alpha = gamma_alpha
        self.tau_alpha = tau_alpha
        self.tau_mu = tau_mu
        self.tau_eta = tau_eta
        self.scramble = scramble
        self.delta = min(delta, self.m)
        self.upd_m = upd_m
        self.upd_k = upd_k
        self.beta_k = beta_k
        self.psi_fn = psi_fn
        self.debug = False  # Temporarily used
        self.reset_memories = True

        # Creating keys (self.K) and memories (self.M)
        self.M = torch.nn.Parameter(torch.empty((self.q, self.m, self.u), dtype=torch.float32))
        if self.upd_k == "ad_hoc_WTA":
            self.register_buffer('K', torch.zeros((self.q, self.m, self.d)))
        else:
            self.K = torch.nn.Parameter(torch.empty((self.q, self.m, self.d), dtype=torch.float32))

        # Buffers for ad_hoc_WTA key updates (average usefulness register buffer "mu" and age "eta")
        if self.upd_k == "ad_hoc_WTA":
            self.register_buffer('mu', torch.zeros(self.q, m, dtype=torch.float))
            self.register_buffer('eta', torch.ones((self.q, m), dtype=torch.float) * self.tau_eta)
            if self.debug:
                self.register_buffer('key_counter', torch.zeros(self.q, m, dtype=torch.float))
        else:
            self.mu = None
            self.eta = None

        # Scrambling stats
        self.register_buffer('scrambling_count', torch.zeros(self.q, dtype=torch.long))

        # Initializing memories and keys
        self.reset_parameters()

    def reset_parameters(self):
        self.__reset_keys()
        if self.reset_memories:
            self.__reset_memories()
        self.__reset_counters()

    def compute_weights(self, x):

        # Shortcuts (notice that "self.delta" is called "k" in shortcuts, while "self.delta-1" is called "z")
        q, m, u, d, k = self.q, self.m, self.u, self.d, self.delta
        b = x.shape[0]
        M_qmu = self.M

        # Ensuring keys are normalized (not needed with ad_hoc_WTA updates)
        if self.upd_k != 'ad_hoc_WTA':
            self.__normalize_keys()
        else:
            x = x.detach()  # In ad-hoc WTA, no gradient is propagated to the layers below through key-matching

        # Mapping the input to the key space using the psi function
        x_bd = psi(x, self.psi_fn, key_size=d, normalize=True)

        # Finding the top responses and indices for the attention procedure
        top_responses_bqk, top_indices_bqk = self.__top_k_attention(x_bd)

        # Probabilities
        top_alpha_bqk = torch.softmax((self.gamma_alpha / math.sqrt(d)) * top_responses_bqk, dim=2)

        if self.debug:

            # Getting the top-1 indices for the current mini-batch
            top1_indices_qb = top_indices_bqk[..., 0].t()
            self.key_counter.data.scatter_add_(dim=1,
                                               index=top1_indices_qb,
                                               src=torch.ones_like(top1_indices_qb, dtype=self.key_counter.dtype))

        # Updating keys with the ad-hoc scheme (also refreshing top-stuff: responses, indices, alpha)
        if self.training and self.upd_k == 'ad_hoc_WTA':
            top_responses_bqk, top_indices_bqk, top_alpha_bqk = \
                self.__update_keys_and_counters(x_bd, top_responses_bqk, top_indices_bqk, top_alpha_bqk)

        # Reading memories and blending them
        if self.upd_m is None:

            # Preparing to read memory units and to blend them
            M_exp_bqmu = M_qmu.view(1, q, m, u).expand(b, q, m, u)

            # Getting top memory units
            top_M_bqku = torch.gather(M_exp_bqmu, dim=2,
                                      index=top_indices_bqk.view(b, q, k, 1).expand(b, q, k, u))

            # Mixing memory units by attention scores
            # -> top_alpha_bqk: [b,q,k], that we un-squeeze to [b,q,1,k]
            # -> top_M_bqku: [b,q,k,u]
            # -> W_bqu: matmul([(b,q),1,k], [(b,q),k,u]) = [b,q,1,u] that we squeeze to [b,q,u]
            W_bqu = torch.matmul(top_alpha_bqk.view(b, q, 1, k), top_M_bqku).squeeze(2)

        elif self.upd_m == 'WTA':

            # Preparing to read memory units and to blend them
            M_exp_bqmu = M_qmu.view(1, q, m, u).expand(b, q, m, u)

            # Dealing with top-1 stuff
            top1_M_exp_bq1u = torch.gather(M_exp_bqmu, dim=2,
                                           index=top_indices_bqk[..., 0:1].view(b, q, 1, 1).expand(b, q, 1, u))

            # Mixing memory units by attention scores
            # -> top1_alpha_bqk: [b,q,k], that we select to [b,k,1] un-squeeze to [b,q,1,1]
            # -> top1_M_exp_bq1u: [b,q,1,u]
            # -> W_bqu: [b,q,1,1] * [b,q,1,u] = [b,q,1,u] that we squeeze to [b,q,u]
            top1_W_bqu = (top_alpha_bqk[..., 0:1].view(b, q, 1, 1) * top1_M_exp_bq1u).squeeze(2)

            # Dealing with top-2-and-following stuff
            top2on_M_exp_bqzu = torch.gather(M_exp_bqmu.detach(), dim=2,
                                             index=top_indices_bqk[..., 1:].view(b, q, k-1, 1).expand(b, q, k-1, u))
            top2on_alpha_bqz = top_alpha_bqk[:, :, 1:]
            if self.upd_k == 'grad_WTA':
                top2on_alpha_bqz = top2on_alpha_bqz.detach()

            # Mixing memory units by attention scores
            # -> top2on_alpha_bqz: [b,q,k-1], that we un-squeeze to [b,q,1,k-1]
            # -> top2on_M_exp_bqzu: [b,q,k-1,u]
            # -> W_bqu: matmul([(b,q),1,k-1], [(b,q),k-1,u]) = [b,q,1,u] that we squeeze to [b,q,u]
            top2on_W_bqu = torch.matmul(top2on_alpha_bqz.view(b, q, 1, k-1), top2on_M_exp_bqzu).squeeze(2)

            # Merging top1 and top-2-and-following stuff
            W_bqu = top1_W_bqu + top2on_W_bqu

        else:

            # What is going on?
            raise NotImplementedError

        return W_bqu

    def forward(self, x):
        raise NotImplementedError

    @torch.no_grad()
    def __reset_memories(self):
        bound = 1. / math.sqrt(self.u)
        torch.nn.init.uniform_(self.M, -bound, bound)

    @torch.no_grad()
    def __reset_keys(self):
        bound = 1. / math.sqrt(self.d)
        torch.nn.init.uniform_(self.K, -bound, bound)
        self.K.data = F.normalize(self.K, p=2.0, dim=2, eps=1e-12, out=None)

    def __reset_counters(self):
        if self.mu is not None:
            self.mu = torch.zeros_like(self.mu)
        if self.eta is not None:
            self.eta.fill_(self.tau_eta)

    def reset_counter(self):
        if self.debug:
            self.key_counter.data = torch.zeros_like(self.key_counter)

    def __top_k_attention(self, x_bd):

        # Matmul([b,d], [d,qm]) = [b,qm], then reshaped (view) to [b,q,m]
        responses_bqm = torch.matmul(x_bd, self.K.view(self.q * self.m, self.d).t()).view(-1, self.q, self.m)
        top_responses_bqk, top_indices_bqk = torch.topk(responses_bqm, k=self.delta, dim=2, largest=True, sorted=True)
        return top_responses_bqk, top_indices_bqk

    @torch.no_grad()
    def __normalize_keys(self, ids=None):
        """
        :param ids: None or a vector with "self.q" elements, with the indices of the keys to consider for each neuron
        """
        if ids is not None:
            ids_exp_q1d = ids.view(self.q, 1, 1).expand(self.q, 1, self.d)
            keys_q1d = torch.gather(self.K, dim=1, index=ids_exp_q1d)
            keys_q1d = F.normalize(keys_q1d, p=2.0, dim=2, eps=1e-12, out=None)
            self.K.scatter_(dim=1, index=ids_exp_q1d, src=keys_q1d)
            return keys_q1d
        else:
            self.K.data = F.normalize(self.K, p=2.0, dim=2, eps=1e-12, out=None)
            return self.K

    def __update_keys_and_counters(self, x_bd, top_responses_bqk, top_indices_bqk, top_alphas_bqk):

        # Saving some shortcuts, notice that "self.delta" is called "k" here
        b, q, d, m, k, u = x_bd.shape[0], self.q, self.d, self.m, self.delta, self.u
        K_qmd = self.K
        mu_qm, eta_qm = self.mu, self.eta

        # Getting the top-1 indices for the current mini-batch
        top1_indices_qb = top_indices_bqk[..., 0].t()

        # Temporarily used
        if self.debug:
            K_initial_qmd = self.K.clone()
            M_initial_qmd = self.M.clone()
            mu_initial_qm = self.mu.clone()
            eta_initial_qm = self.eta.clone()
            top1_indices_initial_qb = top1_indices_qb.clone()
        else:
            K_initial_qmd, M_initial_qmd, mu_initial_qm = None, None, None
            eta_initial_qm, top1_indices_initial_qb = None, None

        # Determining if we need to scramble keys and memories (if scrambling is enabled)
        # (up to one key/memory per neuron, even when the batch size is greater than one)
        if self.scramble:

            # Computing a boolean mask that tells what neurons should be subject to scrambling (scramble_q),
            # a boolean mask associated with the weak keys,
            # and the indices of the elements of x_bd (batch elements) that should replace the
            # scrambled keys (weak_batch_elements_q)
            scramble_q, weak_keys_mask_qm, weak_batch_elements_q = \
                self.__evaluate_scrambling_conditions(top_responses_bqk)

            # Temporarily used
            if self.debug:
                self.__debug_pre_scrambling(top_indices_bqk, top_alphas_bqk, scramble_q,
                                            weak_keys_mask_qm, weak_batch_elements_q, x_bd)

            # If at least one neuron requires a scrambling operation, we do scramble! (in-place)
            if torch.any(scramble_q):
                self.__scramble(x_bd, scramble_q, weak_keys_mask_qm, weak_batch_elements_q, top1_indices_qb)

            # Temporarily used
            if self.debug:
                self.__debug_changes_with_respect_to(K_initial_qmd, M_initial_qmd, mu_initial_qm, eta_initial_qm,
                                                     top1_indices_initial_qb, top1_indices_qb,
                                                     msg="Right after scrambling...")

        # Computing variations to apply to winning keys, eventually using an adaptive learning rate
        key_variations_qbd = (self.beta_k * x_bd).view(1, b, d).expand(q, b, d)  # Delta to add to each winning key

        # Updating the winning keys (one winning key per neuron, for each batch example)
        K_qmd.scatter_add_(dim=1,
                           index=top1_indices_qb.view(q, b, 1).expand(q, b, d),
                           src=key_variations_qbd)  # Adding variations to winning keys

        # Recomputing or updating responses (in the case of batch size > 1, we recompute them all)
        top_responses_bqk, top_indices_bqk, top1_indices_qb = \
            self.__update_top_k_attention(x_bd, top_responses_bqk, top_indices_bqk, top1_indices_qb)

        # Recomputing the softmax over the (now updated) top responses
        top_alphas_bqk = torch.softmax((self.gamma_alpha / math.sqrt(d)) * top_responses_bqk, dim=2)  # [b,q,k]

        # Resetting ages of winning keys
        eta_qm.scatter_(1, top1_indices_qb, 0.)

        # Updating counters: usages ("mu") for the winning keys and ages ("eta") for all the keys
        mu_qm.scatter_add_(dim=1,
                           index=top1_indices_qb,
                           src=torch.ones_like(top1_indices_qb, dtype=mu_qm.dtype))  # Winning keys
        eta_qm += b  # All the keys

        # Temporarily used
        if self.debug:
            self.__debug_changes_with_respect_to(K_initial_qmd, M_initial_qmd, mu_initial_qm, eta_initial_qm,
                                                 top1_indices_initial_qb, top1_indices_qb,
                                                 msg="At the end of the whole key-and-counters update procedure...")

        return top_responses_bqk, top_indices_bqk, top_alphas_bqk

    def __evaluate_scrambling_conditions(self, top_responses_bqk):

        # Shortcuts
        mu_qm = self.mu
        eta_qm = self.eta

        # Computing the max of alphas (we take the smallest of them in case of batch sizes greater than one)
        top1_responses_bq = top_responses_bqk[..., 0]  # Max of responses
        max_of_responses_q, weak_batch_elements_q = torch.min(top1_responses_bq, dim=0)

        # Finding the weak keys, if any (boolean mask)
        weak_keys_mask_qm = torch.logical_and(mu_qm < self.tau_mu, eta_qm >= self.tau_eta)

        # Determining on what neurons scrambling should be applied (boolean mask)
        scramble_q = torch.logical_and(max_of_responses_q < self.tau_alpha, torch.any(weak_keys_mask_qm, dim=1))

        return scramble_q, weak_keys_mask_qm, weak_batch_elements_q

    def __scramble(self, x_bd, scramble_q, weak_keys_mask_qm, weak_batch_elements_q, top1_indices_qb):

        # Shortcuts
        q, d, u = self.q, self.d, self.u
        K_qmd, M_qmu, mu_qm, eta_qm = self.K, self.M, self.mu, self.eta

        # Stats
        self.scrambling_count[scramble_q] += 1

        # Finding the indices of the candidate keys to scramble (one per neuron)
        scramble_candidates_keys_q = torch.max(eta_qm * weak_keys_mask_qm.to(torch.float), dim=1)[1]

        # Neurons that must and must-not be subject to scrambling operations
        scramble_q = scramble_q.to(torch.float)
        no_scramble_q = torch.logical_not(scramble_q).to(torch.float)  # Boolean mask

        # Scrambling keys
        scramble_candidates_keys_exp_q1d = scramble_candidates_keys_q.view(q, 1, 1).expand(q, 1, d)
        scramble_q11 = scramble_q.view(q, 1, 1)
        no_scramble_q11 = no_scramble_q.view(q, 1, 1)

        new_keys_q1d = x_bd.gather(dim=0, index=weak_batch_elements_q.view(q, 1).expand(q, d)).view(q, 1, d)
        old_keys_q1d = K_qmd.gather(dim=1, index=scramble_candidates_keys_exp_q1d)

        K_qmd.scatter_(dim=1,
                       index=scramble_candidates_keys_exp_q1d,
                       src=new_keys_q1d * scramble_q11 + old_keys_q1d * no_scramble_q11)

        # Scrambling memories
        with torch.no_grad():
            scramble_candidates_keys_exp_q1u = scramble_candidates_keys_q.view(q, 1, 1).expand(q, 1, u)
            weak_keys_q1 = top1_indices_qb.gather(dim=1, index=weak_batch_elements_q.view(q, 1))
            weak_keys_exp_q1u = weak_keys_q1.view(q, 1, 1).expand(q, 1, u)

            new_memories_q1u = M_qmu.gather(dim=1, index=weak_keys_exp_q1u)
            old_memories_q1u = M_qmu.gather(dim=1, index=scramble_candidates_keys_exp_q1u)

            M_qmu.scatter_(
                dim=1,
                index=scramble_candidates_keys_exp_q1u,
                src=new_memories_q1u * scramble_q11 + old_memories_q1u * no_scramble_q11)

        # Updating indices of the winning keys (in-place!)
        weak_batch_elements_q1 = weak_batch_elements_q.view(q, 1)
        scramble_q1 = scramble_q.view(q, 1)
        no_scramble_q1 = no_scramble_q.view(q, 1)

        new_top1_indices_q1 = scramble_candidates_keys_q.view(q, 1)
        old_top1_indices_q1 = top1_indices_qb.gather(dim=1, index=weak_batch_elements_q1)

        top1_indices_qb.scatter_(
            dim=1, index=weak_batch_elements_q1,
            src=(new_top1_indices_q1 * scramble_q1 + old_top1_indices_q1 * no_scramble_q1).to(torch.long))

        # Resetting to zero the usage counts ("mu") for keys that were scrambled
        scramble_candidates_keys_q1 = scramble_candidates_keys_q.view(q, 1)

        new_values = 0.
        old_values_q1 = mu_qm.gather(dim=1, index=scramble_candidates_keys_q1)

        mu_qm.scatter_(
            dim=1,
            index=scramble_candidates_keys_q1,
            src=new_values * scramble_q1 + old_values_q1 * no_scramble_q1)

    def __update_top_k_attention(self, x_bd, top_responses_bqk, top_indices_bqk, top1_indices_qb):
        q, d, b = self.q, self.d, x_bd.shape[0]

        if b == 1:

            # Normalizing the winning keys (recall that b=1 here)
            normalized_winning_keys_q1d = self.__normalize_keys(ids=top1_indices_qb)  # Recall that b=1 here

            # Updating responses (the response with the updated key is recomputed, for each neuron)
            response_winning_bq1 = torch.matmul(x_bd, normalized_winning_keys_q1d.view(q, d).t()).view(b, q, 1)
            top_responses_bqk[:, :, 0] = response_winning_bq1.squeeze()

        elif b > 1:

            # Normalizing all the keys (for simplicity - with large batch sizes it is likely easier/faster)
            self.__normalize_keys()

            # Recomputing all responses, re-determining the top-responses
            top_responses_bqk, top_indices_bqk = self.__top_k_attention(x_bd)

            # Re-transposing top-1 indices
            top1_indices_qb = top_indices_bqk[..., 0].t()

        return top_responses_bqk, top_indices_bqk, top1_indices_qb

    def __str__(self):
        s = "- q = " + str(self.q) + "\n"
        s += "- d = " + str(self.d) + "\n"
        s += "- m = " + str(self.m) + "\n"
        s += "- u = " + str(self.u) + "\n"
        s += "- delta = " + str(self.delta) + "\n"
        s += "- gamma_alpha = " + str(self.gamma_alpha) + "\n"
        s += "- tau_alpha = " + str(self.tau_alpha) + "\n"
        s += "- tau_mu = " + str(self.tau_mu) + "\n"
        s += "- tau_eta = " + str(self.tau_eta) + "\n"
        s += "- upd_m = " + str(self.upd_m) + "\n"
        s += "- upd_k = " + str(self.upd_k) + "\n"
        s += "- beta_k = " + str(self.beta_k) + "\n"
        s += "- psi_fn = " + self.psi_fn + "\n"
        s += "- scramble = " + str(self.scramble)
        return s

    def __debug_pre_scrambling(self, top_indices_bqk, top_alphas_bqk, scramble_q,
                               weak_keys_mask_qm, weak_batch_elements_q, x_bd):
        b, q, k, d, m = top_indices_bqk.shape[0], self.q, self.delta, self.d, self.m
        mu_qm, eta_qm = self.mu, self.eta

        # Finding the indices of the candidate keys to scramble (one per neuron)
        scramble_candidates_keys_q = torch.max(eta_qm * weak_keys_mask_qm.to(torch.float), dim=1)[1]

        print("*** __debug_pre_scrambling ***")
        print("No scrambling is applied at all...not yet...")
        s = ""
        torch.set_printoptions(profile='full', linewidth=2000)
        for i in range(0, b):
            if i > 0:
                s += "\n"
            for j in range(0, q):
                if j > 0:
                    s += "\n"
                s += "[batch element " + str(i)
                s += ", neuron " + str(j) + "] " if q > 1 else "]"
                s += "\n- tau_alpha, tau_mu, tau_eta: " + str(self.tau_alpha) + ", "  \
                     + str(self.tau_mu) + ", " + str(self.tau_eta)
                s += "\n- mu:       " + str(mu_qm[j, :])
                s += "\n- eta:      " + str(eta_qm[j, :])
                s += "\n- top-keys (alphas): "
                for a in range(0, k):
                    s += str(top_indices_bqk[i, j, a].item())
                    s += " ({0:.3g})".format(top_alphas_bqk[i, j, a].item())
                    if a < k - 1:
                        s += ", "
                s += "\n- scramble? " + str(scramble_q[j].item())
                if scramble_q[j].item() is True:
                    s += "\n  - what key? " + str(scramble_candidates_keys_q[j].item())
                    s += "\n  - with what batch element? " + str(weak_batch_elements_q[j].item())
                    s += "\n  - data of such element? "
                    s += str(x_bd[weak_batch_elements_q[j], 0:min(3, d)])
                    if d > 3:
                        s += " ~printing only the first 3 components"
        print(s)
        torch.set_printoptions(profile='default')

    def __debug_changes_with_respect_to(self, K_initial_qmd, M_initial_qmd, mu_initial_qm, eta_initial_qm,
                                        top1_indices_initial_qb, top1_indices_qb, msg=None):
        q, m, b, u, d = self.q, self.m, top1_indices_initial_qb.shape[1], self.u, self.d

        changed_keys_qm = torch.greater(torch.max(torch.abs(self.K - K_initial_qmd), dim=2)[0], 1e-5)
        changed_memories_qm = torch.greater(torch.max(torch.abs(self.M - M_initial_qmd), dim=2)[0], 0.)
        changed_top1_indices_qb = torch.greater(torch.abs(top1_indices_initial_qb.to(torch.float) -
                                                          top1_indices_qb.to(torch.float)), 0.)
        changed_mus_qm = torch.greater(torch.abs(self.mu - mu_initial_qm), 0.)
        changed_etas_qm = torch.greater(torch.abs(self.eta - eta_initial_qm), 0.)

        print("*** __debug_changes_with_respect_to ***")
        if msg is not None:
            print(msg)
        s = ""
        for j in range(0, q):
            if j > 0:
                s += "\n"
            s += "[neuron " + str(j) + "]"
            num_changed_keys = torch.sum(changed_keys_qm[j, :]).item()
            num_changed_memories = torch.sum(changed_memories_qm[j, :]).item()
            num_changed_top1_indices = torch.sum(changed_top1_indices_qb[j, :]).item()
            num_changed_mus = torch.sum(changed_mus_qm[j, :]).item()
            num_changed_etas = torch.sum(changed_etas_qm[j, :]).item()
            s += "\n- #changed keys: " + str(num_changed_keys)
            if num_changed_keys > 0:
                if d > 3:
                    s += " ~printing only the first 3 components"
                for t in range(0, m):
                    if changed_keys_qm[j, t].item() is True:
                        s += "\n  - key " + str(t) + ": " + str(self.K[j, t, 0:min(3, d)])
                        s += " (it was: " + str(K_initial_qmd[j, t, 0:min(3, d)]) + ")"
            s += "\n- #changed memory units: " + str(num_changed_memories)
            if num_changed_memories > 0:
                if u > 3:
                    s += " ~printing only the first 3 components"
                for t in range(0, m):
                    if changed_memories_qm[j, t].item() is True:
                        s += "\n  - mem " + str(t) + ": " + str(self.M[j, t, 0:min(3, u)])
                        s += " (it was: " + str(M_initial_qmd[j, t, 0:min(3, u)]) + ")"
            s += "\n- #changed top1 indices: " + str(num_changed_top1_indices)
            if num_changed_top1_indices > 0:
                for i in range(0, b):
                    if changed_top1_indices_qb[j, i].item() is True:
                        s += "\n  - batch element " + str(i) + ": " + str(top1_indices_qb[j, i].item())
                        s += " (it was: " + str(top1_indices_initial_qb[j, i].item()) + ")"
            s += "\n- #changed mus: " + str(num_changed_mus)
            if num_changed_mus > 0:
                for t in range(0, m):
                    if changed_mus_qm[j, t].item() is True:
                        s += "\n  - mu " + str(t) + ": " + str(self.mu[j, t].item())
                        s += " (it was: " + str(mu_initial_qm[j, t].item()) + ")"
            s += "\n- #changed etas: " + str(num_changed_etas)
            if num_changed_etas > 0:
                for t in range(0, m):
                    if changed_etas_qm[j, t].item() is True:
                        s += "\n  - eta " + str(t) + ": " + str(self.eta[j, t].item())
                        s += " (it was: " + str(eta_initial_qm[j, t].item()) + ")"
        print(s)



"""
       █████  █████ ██████   █████           █████ █████   █████ ██████████ ███████████    █████████  ██████████
      ░░███  ░░███ ░░██████ ░░███           ░░███ ░░███   ░░███ ░░███░░░░░█░░███░░░░░███  ███░░░░░███░░███░░░░░█
       ░███   ░███  ░███░███ ░███   ██████   ░███  ░███    ░███  ░███  █ ░  ░███    ░███ ░███    ░░░  ░███  █ ░ 
       ░███   ░███  ░███░░███░███  ░░░░░███  ░███  ░███    ░███  ░██████    ░██████████  ░░█████████  ░██████   
       ░███   ░███  ░███ ░░██████   ███████  ░███  ░░███   ███   ░███░░█    ░███░░░░░███  ░░░░░░░░███ ░███░░█   
       ░███   ░███  ░███  ░░█████  ███░░███  ░███   ░░░█████░    ░███ ░   █ ░███    ░███  ███    ░███ ░███ ░   █
       ░░████████   █████  ░░█████░░████████ █████    ░░███      ██████████ █████   █████░░█████████  ██████████
        ░░░░░░░░   ░░░░░    ░░░░░  ░░░░░░░░ ░░░░░      ░░░      ░░░░░░░░░░ ░░░░░   ░░░░░  ░░░░░░░░░  ░░░░░░░░░░ 
                 A Collectionless AI Project (https://collectionless.ai)
                 Registration/Login: https://unaiverse.io
                 Code Repositories:  https://github.com/collectionlessai/
                 Main Developers:    Stefano Melacci (Project Leader), Christian Di Maio, Tommaso Guidi
"""
from . import messages
from . import p2p
from . import golibp2p
from . import lib_types
import os
import sys
import glob
import json
import ctypes
import hashlib
import platform
import warnings
from typing import cast
from .messages import Msg
from .p2p import P2P, P2PError
from .golibp2p import GoLibP2P  # Your stub interface definition
from .lib_types import TypeInterface  # Assuming TypeInterface handles the void* results


# --- Helper Functions ---
def _get_lib_filename():
    """Determines the correct shared library filename based on the OS."""
    system = platform.system()
    if system == 'Linux':
        return 'unailib*.so'
    elif system == 'Darwin':
        return 'unailib*.dylib'
    elif system == 'Windows':
        return 'unailib*.pyd'
    # This error should ideally never be reached if setup.py ran correctly.
    raise ImportError(f"Unsupported operating system for shared library: {system}")

def _get_file_hash(filepath):
    """Calculates the SHA256 hash of a file, returning None if it doesn't exist."""
    if not os.path.exists(filepath):
        return None
    sha256_hash = hashlib.sha256()
    with open(filepath, "rb") as f:
        for byte_block in iter(lambda: f.read(4096), b""):
            sha256_hash.update(byte_block)
    return sha256_hash.hexdigest()

def _developer_source_check():
    """
    If source files are present (i.e., in a dev environment), check if the
    compiled library is in sync with the source code.
    """
    go_source_file = os.path.join(_lib_dir, 'lib.go')
    hash_file = go_source_file + '.sha256'
    
    # This check only runs if all dev files are present. For users who
    # installed from a wheel, these files won't exist, and this is skipped.
    if os.path.exists(go_source_file) and os.path.exists(hash_file):
        current_source_hash = _get_file_hash(go_source_file)
        
        stored_build_hash = None
        with open(hash_file, 'r') as f:
            stored_build_hash = f.read().strip()
            
        if current_source_hash != stored_build_hash:
            # Use warnings.warn for a standard, non-intrusive developer warning.
            warnings.warn(
                "\n" + "="*80 +
                "\nWARNING: The Go source file (lib.go) has been modified since the shared\n"
                "library was last compiled. Your running code may not reflect recent changes.\n\n"
                "To fix this, run: pip install -e .\n" +
                "="*80,
                UserWarning
            )

# --- Main Library Loading ---
# The shared library is guaranteed by the build process to be in this directory.
_lib_dir = os.path.dirname(os.path.abspath(__file__))
_shared_lib = None

try:
    _results = glob.glob(os.path.join(_lib_dir, _get_lib_filename()))
    _lib_path = os.path.join(_lib_dir, _results[0])
    _shared_lib = ctypes.CDLL(_lib_path)
except OSError as e:
    print(
        f"FATAL: Could not load the required p2p shared library from {_lib_path}.\n"
        "This indicates a corrupted or missing installation. "
        "Please try reinstalling the 'unaiverse' package.\n"
        f"Underlying error: {e}",
        file=sys.stderr
    )
    raise ImportError("Failed to load the UNaIVERSE p2p shared library.") from e

# Run the check after successfully loading the library.
if _shared_lib is not None:
    _developer_source_check()
    print(f"UNaIVERSE: Successfully loaded p2p library from {_lib_path}")

# --- Function Prototypes (argtypes and restype) ---
# Using void* for returned C strings, requiring TypeInterface for conversion/freeing.

# Define argtypes for the Go init function here
_shared_lib.InitializeLibrary.argtypes = [ctypes.c_int, ctypes.c_int, ctypes.c_int, ctypes.c_int, ctypes.c_int]
_shared_lib.InitializeLibrary.restype = None

# Node Lifecycle & Info
_shared_lib.CreateNode.argtypes = [ctypes.c_int, ctypes.c_char_p, ctypes.c_int, ctypes.c_char_p, ctypes.c_int, ctypes.c_int,
                                   ctypes.c_int, ctypes.c_int, ctypes.c_int, ctypes.c_char_p, ctypes.c_char_p, ctypes.c_char_p]
_shared_lib.CreateNode.restype = ctypes.c_void_p  # Treat returned *C.char as opaque pointer

_shared_lib.CloseNode.argtypes = [ctypes.c_int]
_shared_lib.CloseNode.restype = ctypes.c_void_p  # Treat returned *C.char as opaque pointer

_shared_lib.GetNodeAddresses.argtypes = [ctypes.c_int, ctypes.c_char_p]  # Input is still a Python string -> C string
_shared_lib.GetNodeAddresses.restype = ctypes.c_void_p  # Treat returned *C.char as opaque pointer

_shared_lib.GetConnectedPeers.argtypes = [ctypes.c_int]
_shared_lib.GetConnectedPeers.restype = ctypes.c_void_p  # Treat returned *C.char as opaque pointer

_shared_lib.GetRendezvousPeers.argtypes = [ctypes.c_int]
_shared_lib.GetRendezvousPeers.restype = ctypes.c_void_p  # Treat returned *C.char as opaque pointer

# Peer Connection
_shared_lib.ConnectTo.argtypes = [ctypes.c_int, ctypes.c_char_p]
_shared_lib.ConnectTo.restype = ctypes.c_void_p  # Treat returned *C.char as opaque pointer

_shared_lib.DisconnectFrom.argtypes = [ctypes.c_int, ctypes.c_char_p]
_shared_lib.DisconnectFrom.restype = ctypes.c_void_p  # Treat returned *C.char as opaque pointer

# Direct Messaging
_shared_lib.SendMessageToPeer.argtypes = [ctypes.c_int, ctypes.c_char_p, ctypes.c_char_p, ctypes.c_int]
_shared_lib.SendMessageToPeer.restype = ctypes.c_void_p  # Returns status code, not pointer

# Message Queue
_shared_lib.MessageQueueLength.argtypes = [ctypes.c_int]
_shared_lib.MessageQueueLength.restype = ctypes.c_int  # Returns length, not pointer

_shared_lib.PopMessages.argtypes = [ctypes.c_int]
_shared_lib.PopMessages.restype = ctypes.c_void_p  # Treat returned *C.char as opaque pointer

# PubSub
_shared_lib.SubscribeToTopic.argtypes = [ctypes.c_int, ctypes.c_char_p]
_shared_lib.SubscribeToTopic.restype = ctypes.c_void_p  # Treat returned *C.char as opaque pointer

_shared_lib.UnsubscribeFromTopic.argtypes = [ctypes.c_int, ctypes.c_char_p]
_shared_lib.UnsubscribeFromTopic.restype = ctypes.c_void_p  # Treat returned *C.char as opaque pointer

# Relay Client
_shared_lib.ReserveOnRelay.argtypes = [ctypes.c_int, ctypes.c_char_p]
_shared_lib.ReserveOnRelay.restype = ctypes.c_void_p  # Treat returned *C.char as opaque pointer

# Memory Management
# FreeString now accepts the opaque pointer directly
_shared_lib.FreeString.argtypes = [ctypes.c_void_p]
_shared_lib.FreeString.restype = None  # Void return

_shared_lib.FreeInt.argtypes = [ctypes.POINTER(ctypes.c_int)]  # Still expects a pointer to int
_shared_lib.FreeInt.restype = None  # Void return

# --- Python Interface Setup ---

# Import necessary components
# IMPORTANT: TypeInterface (or equivalent logic) MUST now handle converting
# the c_char_p results back to strings/JSON before freeing.
# Ensure TypeInterface methods like from_go_string_to_json are adapted for this.

# Import the stub type for type checking
try:
    from .golibp2p import GoLibP2P  # Your stub interface definition
except ImportError:
    print("Warning: GoLibP2P stub not found. Type checking will be limited.", file=sys.stderr)
    GoLibP2P = ctypes.CDLL

# Cast the loaded library object to the stub type
_shared_lib_typed = cast(GoLibP2P, _shared_lib)

# Attach the typed shared library object to the P2P class
P2P.libp2p = _shared_lib_typed
TypeInterface.libp2p = _shared_lib_typed  # Attach to TypeInterface if needed

# Attach the typed shared library object to the P2PError class

# Define the public API of this package
__all__ = [
    "P2P",
    "P2PError",
    "TypeInterface"  # Expose TypeInterface if users need its conversion helpers directly
]



"""
       █████  █████ ██████   █████           █████ █████   █████ ██████████ ███████████    █████████  ██████████
      ░░███  ░░███ ░░██████ ░░███           ░░███ ░░███   ░░███ ░░███░░░░░█░░███░░░░░███  ███░░░░░███░░███░░░░░█
       ░███   ░███  ░███░███ ░███   ██████   ░███  ░███    ░███  ░███  █ ░  ░███    ░███ ░███    ░░░  ░███  █ ░ 
       ░███   ░███  ░███░░███░███  ░░░░░███  ░███  ░███    ░███  ░██████    ░██████████  ░░█████████  ░██████   
       ░███   ░███  ░███ ░░██████   ███████  ░███  ░░███   ███   ░███░░█    ░███░░░░░███  ░░░░░░░░███ ░███░░█   
       ░███   ░███  ░███  ░░█████  ███░░███  ░███   ░░░█████░    ░███ ░   █ ░███    ░███  ███    ░███ ░███ ░   █
       ░░████████   █████  ░░█████░░████████ █████    ░░███      ██████████ █████   █████░░█████████  ██████████
        ░░░░░░░░   ░░░░░    ░░░░░  ░░░░░░░░ ░░░░░      ░░░      ░░░░░░░░░░ ░░░░░   ░░░░░  ░░░░░░░░░  ░░░░░░░░░░ 
                 A Collectionless AI Project (https://collectionless.ai)
                 Registration/Login: https://unaiverse.io
                 Code Repositories:  https://github.com/collectionlessai/
                 Main Developers:    Stefano Melacci (Project Leader), Christian Di Maio, Tommaso Guidi
"""
import os
import csv
import math
import torch
import random
import pathlib
from PIL import Image
from .clock import Clock
from .dataprops import DataProps
from unaiverse.utils.misc import show_images_grid


class DataStream:
    """
    Base class for handling a generic data stream.
    """

    def __init__(self,
                 props: DataProps,
                 clock: Clock = Clock()) -> None:
        """Initialize a DataStream.

        Args:
            props (DataProps): Properties of the stream.
            clock (Clock): Clock object for time management (usually provided from outside).
        """

        # A stream can be turned off
        self.props = props
        self.clock = clock
        self.data = None
        self.data_timestamp = self.clock.get_time()
        self.data_timestamp_when_got_by = {}
        self.data_tag = -1
        self.data_uuid = None
        self.data_uuid_expected = None
        self.data_uuid_clearable = False
        self.enabled = True

    @staticmethod
    def create(stream: 'DataStream', name: str | None = None, group: str = 'none',
               public: bool = True, pubsub: bool = True):
        """Create and set the name for a given stream, also updates data stream labels.

        Args:
            stream (DataStream): The stream object to modify.
            name (str): The name of the stream.
            group (str): The name of the group to which the stream belongs.
            public (bool): If the stream is going to be served in the public net or the private one.
            pubsub (bool): If the stream is going to be served by broadcasting (PubSub) or not.

        Returns:
            Stream: The modified stream with updated group name.
        """
        assert name is not None or group != 'none', "Must provide either name or group name."
        stream.props.set_group(group)
        if name is not None:
            stream.props.set_name(name)
        else:
            stream.props.set_name(stream.props.get_name() + "@" + group)  # If name is None, then a group was provided
        stream.props.set_public(public)
        stream.props.set_pubsub(pubsub)
        if (stream.props.is_flat_tensor_with_labels() and
                len(stream.props.tensor_labels) == 1 and stream.props.tensor_labels[0] == 'unk'):
            stream.props.tensor_labels[0] = group if group != 'none' else name
        return stream

    def enable(self):
        """Enable the stream, allowing data to be retrieved.

        Returns:
            None
        """
        self.enabled = True

    def disable(self):
        """Disable the stream, preventing data from being retrieved.

        Returns:
            None
        """
        self.enabled = False

    def get_props(self) -> DataProps:
        return self.props

    def net_hash(self, prefix: str):
        return self.get_props().net_hash(prefix)

    @staticmethod
    def peer_id_from_net_hash(net_hash):
        return DataProps.peer_id_from_net_hash(net_hash)

    @staticmethod
    def name_or_group_from_net_hash(net_hash):
        return DataProps.name_or_group_from_net_hash(net_hash)

    @staticmethod
    def is_pubsub_from_net_hash(net_hash):
        return DataProps.is_pubsub_from_net_hash(net_hash)

    def is_pubsub(self):
        return self.get_props().is_pubsub()

    def is_public(self):
        return self.get_props().is_public()

    def __getitem__(self, idx: int) -> tuple[torch.Tensor | None, int]:
        """Get item for a specific clock cycle. Not implemented for base class: it will be implemented in buffered data
        streams or stream that can generate data on-the-fly.

        Args:
            idx (int): Index of the data to retrieve.

        Raises:
            ValueError: Always, since this method should be overridden.
        """
        raise ValueError("Not implemented (expected to be only present in data streams that are buffered or "
                         "that can generated on the fly)")

    def __str__(self) -> str:
        """String representation of the object.

        Returns:
            str: String representation of the object.
        """
        return f"[DataStream] enabled: {self.enabled}\n\tprops: {self.props}"

    def __len__(self):
        """Get the length of the stream.

        Returns:
            int: Infinity (as this is a lifelong stream).
        """
        return math.inf

    def set(self, data: torch.Tensor | Image.Image | str, data_tag: int = -1) -> bool:
        """Set a new data sample into the stream, that will be provided when calling "get()".

        Args:
            data (torch.Tensor): Data sample to set.
            data_tag (int): Custom data time tag >= 0 (Default: -1, meaning no tags).

        Returns:
            bool: True if data was accepted based on time constraints, else False.
        """
        if self.props.delta <= 0. or self.props.delta <= (self.clock.get_time() - self.data_timestamp):
            self.data = self.props.adapt_tensor_to_tensor_labels(data) if data is not None else None
            self.data_timestamp = self.clock.get_cycle_time()
            self.data_tag = data_tag if data_tag >= 0 else self.clock.get_cycle()
            return True
        else:
            return False

    def get(self, requested_by: str | None = None) -> torch.Tensor | None:
        """Get the most recent data sample from the stream (i.e., the last one that was "set").

        Returns:
            torch.Tensor | None: Adapted data sample if available.
        """
        if self.enabled:
            if requested_by is not None:
                if (requested_by not in self.data_timestamp_when_got_by or
                        self.data_timestamp_when_got_by[requested_by] != self.data_timestamp):
                    self.data_timestamp_when_got_by[requested_by] = self.data_timestamp
                    return self.data
                else:
                    return None
            else:
                return self.data
        else:
            return None

    def get_timestamp(self) -> float:
        return self.data_timestamp

    def get_tag(self) -> int:
        return self.data_tag

    def get_uuid(self, expected: bool = False) -> str | None:
        return self.data_uuid if not expected else self.data_uuid_expected

    def set_uuid(self, ref_uuid: str | None, expected: bool = False):
        if not expected:
            self.data_uuid = ref_uuid
        else:
            self.data_uuid_expected = ref_uuid

        # When we "set" a UUID, even if None, we unmark the "clear UUID" flag, to be sure that nobody is going to
        # accidentally clear this information
        self.data_uuid_clearable = False

    def mark_uuid_as_clearable(self):
        self.data_uuid_clearable = True

    def clear_uuid_if_marked_as_clearable(self):
        if self.data_uuid_clearable:
            self.data_uuid = None
            self.data_uuid_expected = None
            self.data_uuid_clearable = False

    def set_props(self, data_stream: 'DataStream'):
        """Set (edit) the data properties picking up the ones of another DataStream.

        Args:
            data_stream (DataStream): the source DataStream from which DataProp is taken.

        Returns:
            None
        """
        self.props = data_stream.props


class BufferedDataStream(DataStream):
    """
    Data stream with buffer support to store historical data.
    """

    def __init__(self, props: DataProps, clock: Clock = Clock(), is_static: bool = False):
        """Initialize a BufferedDataStream.

        Args:
            is_static (bool): If True, the buffer stores only one item that is reused.
        """
        super().__init__(props=props, clock=clock)

        # We store the data samples, and we cache their text representation (for speed)
        self.data_buffer = []
        self.text_buffer = []

        self.is_static = is_static  # A static stream store only one sample and always yields it

        # We need to remember the fist cycle in which we started buffering and the last one we buffered
        self.first_cycle = -1
        self.last_cycle = -1
        self.last_get_cycle = -2  # Keep it to -2 (since -1 is the starting value for cycles)
        self.buffered_data_index = -1

        self.restart_before_next_get = set()

    def get(self, requested_by: str | None = None) -> tuple[torch.Tensor | None, float]:
        """Get the current data sample based on cycle and buffer.

        Returns:
            torch.Tensor | None: Current buffered sample.
        """
        if requested_by is not None and requested_by in self.restart_before_next_get:
            self.restart_before_next_get.remove(requested_by)
            self.restart()

        cycle = self.clock.get_cycle() - self.first_cycle  # This ensures that first get clock = first sample

        # These two lines might make you think "hey, call super().set(self[cycle]), it is the same!"
        # however, it is not like that, since "set" will also call "adapt_to_labels", that is not needed for
        # buffered streams
        if (self.last_get_cycle != cycle and
                (self.props.delta <= 0. or self.props.delta <= (self.clock.get_time() - self.data_timestamp))):
            self.last_get_cycle = cycle
            self.buffered_data_index += 1
            new_data, new_tag = self[self.buffered_data_index]
            self.data = new_data
            self.data_timestamp = self.clock.get_cycle_time()
            self.data_tag = new_tag
        return super().get(requested_by)

    def set(self, data: torch.Tensor, data_tag: int = -1):
        """Store a new data sample into the buffer.

        Args:
            data (torch.Tensor): Data to store.
            data_tag (int): Custom data time tag >= 0 (Default: -1, meaning no tags).

        Returns:
            bool: True if the data was buffered.
        """
        ret = super().set(data, data_tag)

        if ret:
            if not self.is_static or len(self.data_buffer) == 0:
                self.data_buffer.append((self.props.adapt_tensor_to_tensor_labels(data), self.get_tag()))
                if self.props.is_flat_tensor_with_labels():
                    self.text_buffer.append(self.props.to_text(data))
                elif self.props.is_text():
                    self.text_buffer.append(data)

                # Boilerplate
                if self.first_cycle < 0:
                    self.first_cycle = self.clock.get_cycle()
                    self.last_cycle = self.first_cycle
                else:

                    # Filling gaps with "None"
                    cycle = self.clock.get_cycle()
                    if cycle > self.last_cycle + 1:
                        for cycle in range(cycle, self.last_cycle + 1):
                            self.data_buffer.append((None, -1))
                        self.last_cycle = cycle - 1

                    self.last_cycle += 1
        return ret

    def __getitem__(self, idx: int) -> tuple[torch.Tensor | None, int]:
        """Retrieve a sample from the buffer based on the given clock cycle.

        Args:
            idx (int): Index (>=0) of the sample.

        Returns:
            torch.Tensor | None: The sample, if available.
        """
        if not self.is_static:
            if idx >= self.__len__() or idx < 0:
                return None, -1
            data, data_tag = self.data_buffer[idx]
        else:
            data, data_tag = self.data_buffer[0]
        return data, data_tag if data_tag >= 0 else self.clock.get_cycle() - self.first_cycle

    def __len__(self):
        """Get number of samples in the buffer.

        Returns:
            int: Number of buffered samples.
        """
        return len(self.data_buffer)

    def set_first_cycle(self, cycle):
        """Manually set the first cycle for the buffer.

        Args:
            cycle (int): Global cycle to start from.
        """
        self.first_cycle = cycle
        self.last_cycle = cycle + len(self)

    def get_first_cycle(self):
        """Get the first cycle of the stream.

        Returns:
            int: First cycle value.
        """
        return self.first_cycle

    def restart(self):
        """Restart the buffer using the current clock cycle.
        """
        self.set_first_cycle(max(self.clock.get_cycle(), 0))
        self.buffered_data_index = -1
        self.data_timestamp_when_got_by = {}

    def plan_restart_before_next_get(self, requested_by: str):
        self.restart_before_next_get.add(requested_by)

    def clear_buffer(self):
        """Clear the data buffer
        """
        self.data_buffer = []
        self.text_buffer = []

        self.first_cycle = -1
        self.last_cycle = -1
        self.last_get_cycle = -2  # Keep it to -2 (since -1 is the starting value for cycles)
        self.buffered_data_index = -1

        self.restart_before_next_get = set()

    def shuffle_buffer(self, seed: int = -1):
        old_buffer = self.data_buffer
        indices = list(range(len(old_buffer)))

        state = random.getstate()
        if seed >= 0:
            random.seed(seed)
        random.shuffle(indices)
        if seed >= 0:
            random.setstate(state)

        self.data_buffer = []
        k = 0
        for i in indices:
            self.data_buffer.append((old_buffer[i][0], old_buffer[k][1]))
            k += 1

        if self.text_buffer is not None and len(self.text_buffer) == len(self.data_buffer):
            old_text_buffer = self.text_buffer
            self.text_buffer = []
            for i in indices:
                self.text_buffer.append(old_text_buffer[i])

    def to_text_snippet(self, length: int | None = None):
        """Convert buffered text samples to a single long string.

        Args:
            length (int | None): Optional length of the resulting text snippet.

        Returns:
            str | None: Human-readable text sequence.
        """
        if self.text_buffer is not None and len(self.text_buffer) > 0:
            if length is not None:
                le = max(length // 2, 1)
                text = " ".join(self.text_buffer[0:min(le, len(self.text_buffer))])
                text += (" ... " + (" ".join(self.text_buffer[max(le, len(self.text_buffer) - le):]))) \
                    if len(self.text_buffer) > le else ""
            else:
                text = " ".join(self.text_buffer)
        else:
            text = None
        return text

    def get_since_timestamp(self, since_what_timestamp: float, stride: int = 1) -> (
            tuple[list[int] | None, list[torch.Tensor | None] | None, int, DataProps]):
        """Retrieve all samples starting from a given timestamp.

        Args:
            since_what_timestamp (float): Timestamp in seconds.
            stride (int): Sampling stride.

        Returns:
            Tuple containing list of cycles, data, current cycle, and data properties.
        """
        since_what_cycle = self.clock.time2cycle(since_what_timestamp)
        return self.get_since_cycle(since_what_cycle, stride)

    def get_since_cycle(self, since_what_cycle: int, stride: int = 1) -> (
            tuple[list[int] | None, list[torch.Tensor | None] | None, int, DataProps]):
        """Retrieve all samples starting from a given clock cycle.

        Args:
            since_what_cycle (int): Cycle number.
            stride (int): Stride to skip cycles.

        Returns:
            Tuple with cycles, data, current cycle, and properties.
        """
        assert stride >= 1 and isinstance(stride, int), f"Invalid stride: {stride}"

        # Notice: this whole routed never calls ".get()", on purpose! it must be as it is
        global_cycle = self.clock.get_cycle()
        if global_cycle < 0:
            return None, None, -1, self.props

        # Fist check: ensure we do not go beyond the first clock and counting the resulting number of steps
        since_what_cycle = max(since_what_cycle, 0)
        num_steps = global_cycle - since_what_cycle + 1

        # Second check: now we compute the index we should pass to get item
        since_what_idx_in_getitem = since_what_cycle - self.first_cycle

        ret_cycles = []
        ret_data = []

        for k in range(0, num_steps, stride):
            _idx = since_what_idx_in_getitem + k
            _data, _ = self[_idx]

            if _data is not None:
                ret_cycles.append(since_what_cycle + k)
                ret_data.append(_data)

        return ret_cycles, ret_data, global_cycle, self.props


class System(DataStream):
    def __init__(self):
        super().__init__(props=DataProps(name=System.__name__, data_type="text", data_desc="System stream",
                                         pubsub=False))
        self.set("ping")

    def get(self, requested_by: str | None = None):
        return super().get()


class Dataset(BufferedDataStream):
    """
    A buffered dataset that streams data from a PyTorch dataset and simulates data-streams for input/output.
    """

    def __init__(self, tensor_dataset: torch.utils.data.Dataset, shape: tuple, index: int = 0, batch_size: int = 1):
        """Initialize a Dataset instance, which wraps around a PyTorch Dataset.

        Args:
            tensor_dataset (torch.utils.data.Dataset): The PyTorch Dataset to wrap.
            shape (tuple): The shape of each sample from the data stream.
            index (int): The index of the element returned by __getitem__ to pick up.
        """
        sample = tensor_dataset[0][index]
        if isinstance(sample, torch.Tensor):
            dtype = sample.dtype
        elif isinstance(sample, int):
            dtype = torch.long
        elif isinstance(sample, float):
            dtype = torch.float32
        else:
            raise ValueError("Expected tensor data or a scalar")

        super().__init__(props=DataProps(name=Dataset.__name__,
                                         data_type="tensor",
                                         data_desc="dataset",
                                         tensor_shape=shape,
                                         tensor_dtype=dtype,
                                         pubsub=True))

        n = len(tensor_dataset)
        b = batch_size
        nb = math.ceil(float(n) / float(b))
        r = n - b * (nb - 1)

        for i in range(0, nb):
            batch = []
            if i == (nb - 1):
                b = r

            for j in range(0, b):
                sample = tensor_dataset[i * b + j][index]
                if isinstance(sample, (int, float)):
                    sample = torch.tensor(sample, dtype=dtype)
                batch.append(sample)

            self.data_buffer.append((torch.stack(batch), -1))

        # It was buffered previously than every other thing
        self.restart()


class ImageFileStream(BufferedDataStream):
    """
    A buffered dataset for image data.
    """

    def __init__(self, image_dir: str, list_of_image_files: str,
                 device: torch.device = None, circular: bool = True, show_images: bool = False):
        """Initialize an ImageFileStream instance for streaming image data.

        Args:
            image_dir (str): The directory containing image files.
            list_of_image_files (str): Path to the file with list of file names of the images.
            device (torch.device): The device to store the tensors on. Default is CPU.
            circular (bool): Whether to loop the dataset or not. Default is True.
        """
        self.image_dir = image_dir
        self.device = device if device is not None else torch.device("cpu")
        self.circular = circular

        # Reading the image file
        # (assume a file with one filename per line or a CSV format with lines such as: cat.jpg,cat,mammal,animal)
        self.image_paths = []

        # Calling the constructor
        super().__init__(props=DataProps(name=ImageFileStream.__name__,
                                         data_type="img",
                                         pubsub=True))

        with open(list_of_image_files, 'r') as f:
            for line in f:
                parts = line.strip().split(',')  # Tolerates if it is a CVS and the first field is the image file name
                image_name = parts[0]
                self.image_paths.append(os.path.join(image_dir, image_name))

        # It was buffered previously than every other thing
        self.last_cycle = -1
        self.first_cycle = self.last_cycle - len(self.image_paths) + 1

        # Possibly print to screen the "clickable" list of images
        if show_images:
            show_images_grid(self.image_paths)
            for i, image_path in enumerate(self.image_paths):
                abs_path = os.path.abspath(image_path)
                file_url = pathlib.Path(abs_path).as_uri()
                basename = os.path.basename(abs_path)  # 'photo.jpg'
                parent = os.path.basename(os.path.dirname(abs_path))  # 'images'
                label = os.path.join(parent, basename) if parent else basename
                clickable_label = f"\033]8;;{file_url}\033\\[{label}]\033]8;;\033\\"
                print(str(i) + " => " + clickable_label)

    def __len__(self):
        """Return the number of images in the dataset.

        Returns:
            int: Number of images in the dataset.
        """
        return len(self.image_paths)

    def __getitem__(self, idx: int) -> tuple[torch.Tensor | None, int]:
        """Get the image and label for the specified cycle number.

        Args:
            idx (int): The cycle number to retrieve data for.

        Returns:
            tuple: A tuple of tensors (image, label) for the specified cycle.
        """
        if self.circular:
            idx %= self.__len__()
        else:
            if idx >= self.__len__() or idx < 0:
                return None, -1

        image = Image.open(self.image_paths[idx])
        return image, self.clock.get_cycle() - self.first_cycle


class LabelStream(BufferedDataStream):
    """
    A buffered stream for single and multi-label annotations.
    """

    def __init__(self, label_dir: str, label_file_csv: str,
                 device: torch.device = None, circular: bool = True, single_class: bool = False,
                 line_header: bool = False):
        """Initialize an LabelStream instance for streaming labels.

        Args:
            label_dir (str): The directory containing image files.
            label_file_csv (str): Path to the CSV file with labels for the images.
            device (torch.device): The device to store the tensors on. Default is CPU.
            circular (bool): Whether to loop the dataset or not. Default is True.
            single_class (bool): Whether to only consider a single class for labeling. Default is False.
        """
        self.label_dir = label_dir
        self.device = device if device is not None else torch.device("cpu")
        self.circular = circular

        # Reading the label file
        # (assume a file with a labeled element per line or a CSV format with lines such as: cat.jpg,cat,mammal,animal)
        self.labels = []

        class_names = {}
        with open(label_file_csv, 'r') as f:
            for line in f:
                parts = line.strip().split(',')
                label = parts[1:]
                for lab in label:
                    class_names[lab] = True
        class_name_to_index = {}
        class_names = list(class_names.keys())

        # Call the constructor
        super().__init__(props=DataProps(name=LabelStream.__name__,
                                         data_type="tensor",
                                         data_desc="label stream",
                                         tensor_shape=(1, len(class_names)),
                                         tensor_dtype=str(torch.float),
                                         tensor_labels=class_names,
                                         tensor_labeling_rule="geq0.5" if not single_class else "max",
                                         pubsub=True))

        for idx, class_name in enumerate(class_names):
            class_name_to_index[class_name] = idx

        with open(label_file_csv, 'r') as f:
            for line in f:
                parts = line.strip().split(',')
                label = parts if not line_header else parts[1:]
                target_vector = torch.zeros((1, len(class_names)), dtype=torch.float32)
                for lab in label:
                    idx = class_name_to_index[lab]
                    target_vector[0, idx] = 1.
                self.labels.append(target_vector)

        # It was buffered previously than every other thing
        self.last_cycle = -1
        self.first_cycle = self.last_cycle - len(self.labels) + 1

    def __len__(self):
        """Return the number of labels in the dataset.

        Returns:
            int: Number of labels in the dataset.
        """
        return len(self.labels)

    def __getitem__(self, idx: int) -> torch.Tensor | None:
        """Get the image and label for the specified cycle number.

        Args:
            idx (int): The cycle number to retrieve data for.

        Returns:
            tuple: A tuple of tensors (image, label) for the specified cycle.
        """
        if self.circular:
            idx %= self.__len__()
        else:
            if idx >= self.__len__() or idx < 0:
                return None, -1

        label = self.labels[idx].unsqueeze(0).to(self.device)  # Multi-label vector for the image
        return self.props.adapt_tensor_to_tensor_labels(label), self.clock.get_cycle() - self.first_cycle


class TokensStream(BufferedDataStream):
    """
    A buffered dataset for tokenized text, where each token is paired with its corresponding labels.
    """

    def __init__(self, tokens_file_csv: str, circular: bool = True, max_tokens: int = -1):
        """Initialize a Tokens instance for streaming tokenized data and associated labels.

        Args:
            tokens_file_csv (str): Path to the CSV file containing token data.
            circular (bool): Whether to loop the dataset or not. Default is True.
            max_tokens (int): Whether to cut the stream to a maximum number of tokens. Default is -1 (no cut).
        """
        self.circular = circular

        # Reading the data file (assume a token per line or a CSV format with lines such as:
        # token,category_label1,category_label2,etc.)
        tokens = []
        with open(tokens_file_csv, 'r') as f:
            for line in f:
                parts = next(csv.reader([line], quotechar='"', delimiter=','))
                tokens.append(parts[0])
                if 0 < max_tokens <= len(tokens):
                    break

        # Vocabulary
        idx = 0
        word2id = {}
        sorted_stream_of_tokens = sorted(tokens)
        for token in sorted_stream_of_tokens:
            if token not in word2id:
                word2id[token] = idx
                idx += 1
        id2word = [""] * len(word2id)
        for _word, _id in word2id.items():
            id2word[_id] = _word

        # Calling the constructor
        super().__init__(props=DataProps(name=TokensStream.__name__,
                                         data_type="text",
                                         data_desc="stream of words",
                                         stream_to_proc_transforms=word2id,
                                         proc_to_stream_transforms=id2word,
                                         pubsub=True))

        # Tokenized text
        for i, token in enumerate(tokens):
            data = token
            self.data_buffer.append((data, -1))

        # It was buffered previously than every other thing
        self.restart()

    def __getitem__(self, idx: int) -> torch.Tensor | None:
        """Get the image and label for the specified cycle number.

        Args:
            idx (int): The index to retrieve data for.

        Returns:
            tuple: A tuple of tensors (image, label) for the specified cycle.
        """
        if self.circular:
            idx %= self.__len__()
        return super().__getitem__(idx)



"""
       █████  █████ ██████   █████           █████ █████   █████ ██████████ ███████████    █████████  ██████████
      ░░███  ░░███ ░░██████ ░░███           ░░███ ░░███   ░░███ ░░███░░░░░█░░███░░░░░███  ███░░░░░███░░███░░░░░█
       ░███   ░███  ░███░███ ░███   ██████   ░███  ░███    ░███  ░███  █ ░  ░███    ░███ ░███    ░░░  ░███  █ ░ 
       ░███   ░███  ░███░░███░███  ░░░░░███  ░███  ░███    ░███  ░██████    ░██████████  ░░█████████  ░██████   
       ░███   ░███  ░███ ░░██████   ███████  ░███  ░░███   ███   ░███░░█    ░███░░░░░███  ░░░░░░░░███ ░███░░█   
       ░███   ░███  ░███  ░░█████  ███░░███  ░███   ░░░█████░    ░███ ░   █ ░███    ░███  ███    ░███ ░███ ░   █
       ░░████████   █████  ░░█████░░████████ █████    ░░███      ██████████ █████   █████░░█████████  ██████████
        ░░░░░░░░   ░░░░░    ░░░░░  ░░░░░░░░ ░░░░░      ░░░      ░░░░░░░░░░ ░░░░░   ░░░░░  ░░░░░░░░░  ░░░░░░░░░░ 
                 A Collectionless AI Project (https://collectionless.ai)
                 Registration/Login: https://unaiverse.io
                 Code Repositories:  https://github.com/collectionlessai/
                 Main Developers:    Stefano Melacci (Project Leader), Christian Di Maio, Tommaso Guidi
"""
import os
import ast
import sys
import time
import json
import math
import threading
from tqdm import tqdm
from pathlib import Path
from datetime import datetime


class GenException(Exception):
    """Base exception for this application (a simple wrapper around a generic Exception)."""
    pass


def save_node_addresses_to_file(node, dir_path: str, public: bool,
                                filename: str = "addresses.txt", append: bool = False):
    address_file = os.path.join(dir_path, filename)
    with open(address_file, "w" if not append else "a") as file:
        file.write(node.hosted.get_name() + ";" +
                   str(node.get_public_addresses() if public else node.get_world_addresses()) + "\n")
        file.flush()


def get_node_addresses_from_file(dir_path: str, filename: str = "addresses.txt") -> dict[str, list[str]]:
    ret = {}
    with open(os.path.join(dir_path, filename)) as file:
        lines = file.readlines()

        # Old file format
        if lines[0].strip() == "/":
            addresses = []
            for line in lines:
                _line = line.strip()
                if len(_line) > 0:
                    addresses.append(_line)
            ret["unk"] = addresses
            return ret

        # New file format
        for line in lines:
            if line.strip().startswith("***"):  # Header marker
                continue
            comma_separated_values = [v.strip() for v in line.split(';')]
            node_name, addresses_str = comma_separated_values
            ret[node_name] = ast.literal_eval(addresses_str)  # Name appearing multiple times? the last entry is kept

    return ret


class Silent:
    def __init__(self, ignore: bool = False):
        self.ignore = ignore

    def __enter__(self):
        if not self.ignore:
            self._original_stdout = sys.stdout
            sys.stdout = open(os.devnull, "w")

    def __exit__(self, exc_type, exc_val, exc_tb):
        if not self.ignore:
            sys.stdout.close()
            sys.stdout = self._original_stdout


# The countdown function
def countdown_start(seconds: int, msg: str):
    class TqdmPrintRedirector:
        def __init__(self, tqdm_instance):
            self.tqdm_instance = tqdm_instance
            self.original_stdout = sys.__stdout__

        def write(self, s):
            if s.strip():  # Ignore empty lines (needed for the way tqdm works)
                self.tqdm_instance.write(s, file=self.original_stdout)

        def flush(self):
            pass  # Tqdm handles flushing

    def drawing(secs: int, message: str):
        with tqdm(total=secs, desc=message, file=sys.__stdout__) as t:
            sys.stdout = TqdmPrintRedirector(t)  # Redirect prints to tqdm.write
            for i in range(secs):
                time.sleep(1)
                t.update(1.)
            sys.stdout = sys.__stdout__  # Restore original stdout

    sys.stdout.flush()
    handle = threading.Thread(target=drawing, args=(seconds, msg))
    handle.start()
    return handle


def countdown_wait(handle):
    handle.join()


def check_json_start(file: str, msg: str, delete_existing: bool = False):
    from rich.json import JSON
    from rich.console import Console
    cons = Console(file=sys.__stdout__)

    if delete_existing:
        if os.path.exists(file):
            os.remove(file)

    def checking(file_path: str, console: Console):
        print(msg)
        prev_dict = {}
        while True:
            if os.path.exists(file_path):
                try:
                    with open(file_path, "r") as f:
                        json_dict = json.load(f)
                        if json_dict != prev_dict:
                            now = datetime.now()
                            console.print("─" * 80)
                            console.print("Printing updated file "
                                          "(print time: " + now.strftime("%Y-%m-%d %H:%M:%S") + ")")
                            console.print("─" * 80)
                            console.print(JSON.from_data(json_dict))
                        prev_dict = json_dict
                except KeyboardInterrupt:
                    break
                except Exception:
                    pass
            time.sleep(1)

    handle = threading.Thread(target=checking, args=(file, cons), daemon=True)
    handle.start()
    return handle


def check_json_start_wait(handle):
    handle.join()


def show_images_grid(image_paths, max_cols=3):
    import matplotlib.pyplot as plt
    import matplotlib.image as mpimg

    n = len(image_paths)
    cols = min(max_cols, n)
    rows = math.ceil(n / cols)

    # Load images
    images = [mpimg.imread(p) for p in image_paths]

    # Determine figure size based on image sizes
    widths, heights = zip(*[(img.shape[1], img.shape[0]) for img in images])

    # Use average width/height for scaling
    avg_width = sum(widths) / len(widths)
    avg_height = sum(heights) / len(heights)

    fig_width = cols * avg_width / 100
    fig_height = rows * avg_height / 100

    fig, axes = plt.subplots(rows, cols, figsize=(fig_width, fig_height))
    axes = axes.flatten() if n > 1 else [axes]

    fig.canvas.manager.set_window_title("Image Grid")

    # Hide unused axes
    for ax in axes[n:]:
        ax.axis('off')

    for idx, (ax, img) in enumerate(zip(axes, images)):
        ax.imshow(img)
        ax.axis('off')
        ax.set_title(str(idx), fontsize=12, fontweight='bold')

    # Display images
    for ax, img in zip(axes, images):
        ax.imshow(img)
        ax.axis('off')

    plt.subplots_adjust(wspace=0, hspace=0)

    # Turn on interactive mode
    plt.ion()
    plt.show()

    fig.canvas.draw()
    plt.pause(0.1)


class FileTracker:
    def __init__(self, folder, ext=".json"):
        self.folder = Path(folder)
        self.ext = ext.lower()
        self.last_state = self._scan_files()

    def _scan_files(self):
        state = {}
        for file in self.folder.iterdir():
            if file.is_file() and file.suffix.lower() == self.ext:
                state[file.name] = os.path.getmtime(file)
        return state

    def something_changed(self):
        new_state = self._scan_files()
        created = [f for f in new_state if f not in self.last_state]
        modified = [f for f in new_state
                    if f in self.last_state and new_state[f] != self.last_state[f]]
        self.last_state = new_state
        return created or modified


def prepare_key_dir(app_name):
    app_name = app_name.lower()
    if os.name == "nt":  # Windows
        if os.getenv("APPDATA") is not None:
            key_dir = os.path.join(os.getenv("APPDATA"), "Local", app_name)  # Expected
        else:
            key_dir = os.path.join(str(Path.home()), f".{app_name}")  # Fallback
    else:  # Linux/macOS
        key_dir = os.path.join(str(Path.home()), f".{app_name}")
    os.makedirs(key_dir, exist_ok=True)
    return key_dir


def get_key_considering_multiple_sources(key_variable: str | None) -> str:

    # Creating folder (if needed) to store the key
    try:
        key_dir = prepare_key_dir(app_name="UNaIVERSE")
    except Exception:
        raise GenException("Cannot create folder to store the key file")
    key_file = os.path.join(key_dir, "key")

    # Getting from an existing file
    key_from_file = None
    if os.path.exists(key_file):
        with open(key_file, "r") as f:
            key_from_file = f.read().strip()

    # Getting from env variable
    key_from_env = os.getenv("NODE_KEY", None)

    # Getting from code-specified option
    if key_variable is not None and len(key_variable.strip()) > 0:
        key_from_var = key_variable.strip()
        if key_from_var.startswith("<") and key_from_var.endswith(">"):  # Something like <UNAIVERSE_KEY_GOES_HERE>
            key_from_var = None
    else:
        key_from_var = None

    # Finding valid sources and checking if multiple keys were provided
    _keys = [key_from_var, key_from_env, key_from_file]
    _source_names = ["your code", "env variable 'NODE_KEY'", f"cache file {key_file}"]
    source_names = []
    mismatching = False
    multiple_source = False
    first_key = None
    first_source = None
    _prev_key = None
    for i, (_key, _source_name) in enumerate(zip(_keys, _source_names)):
        if _key is not None:
            source_names.append(_source_name)
            if _prev_key is not None:
                if _key != _prev_key:
                    mismatching = True
                multiple_source = True
            else:
                _prev_key = _key
                first_key = _key
                first_source = _source_name

    if len(source_names) > 0:
        msg = ""
        if multiple_source and not mismatching:
            msg = "UNaIVERSE key (the exact same key) present in multiple locations: " + ", ".join(source_names)
        if multiple_source and mismatching:
            msg = "UNaIVERSE keys (different keys) present in multiple locations: " + ", ".join(source_names)
            msg += "\nLoaded the one stored in " + first_source
        if not multiple_source:
            msg = f"UNaIVERSE key loaded from {first_source}"
        print(msg)
        return first_key
    else:

        # If no key present, ask user and save to file
        print("UNaIVERSE key not present in " + ", ".join(_source_names))
        print("If you did not already do it, go to https://unaiverse.io, login, and generate a key")
        key = input("Enter your UNaIVERSE key, that will be saved to the cache file: ").strip()
        with open(key_file, "w") as f:
            f.write(key)
        return key



"""
       █████  █████ ██████   █████           █████ █████   █████ ██████████ ███████████    █████████  ██████████
      ░░███  ░░███ ░░██████ ░░███           ░░███ ░░███   ░░███ ░░███░░░░░█░░███░░░░░███  ███░░░░░███░░███░░░░░█
       ░███   ░███  ░███░███ ░███   ██████   ░███  ░███    ░███  ░███  █ ░  ░███    ░███ ░███    ░░░  ░███  █ ░ 
       ░███   ░███  ░███░░███░███  ░░░░░███  ░███  ░███    ░███  ░██████    ░██████████  ░░█████████  ░██████   
       ░███   ░███  ░███ ░░██████   ███████  ░███  ░░███   ███   ░███░░█    ░███░░░░░███  ░░░░░░░░███ ░███░░█   
       ░███   ░███  ░███  ░░█████  ███░░███  ░███   ░░░█████░    ░███ ░   █ ░███    ░███  ███    ░███ ░███ ░   █
       ░░████████   █████  ░░█████░░████████ █████    ░░███      ██████████ █████   █████░░█████████  ██████████
        ░░░░░░░░   ░░░░░    ░░░░░  ░░░░░░░░ ░░░░░      ░░░      ░░░░░░░░░░ ░░░░░   ░░░░░  ░░░░░░░░░  ░░░░░░░░░░ 
                 A Collectionless AI Project (https://collectionless.ai)
                 Registration/Login: https://unaiverse.io
                 Code Repositories:  https://github.com/collectionlessai/
                 Main Developers:    Stefano Melacci (Project Leader), Christian Di Maio, Tommaso Guidi
"""
import math
from unaiverse.networking.p2p.messages import Msg
from unaiverse.networking.p2p.p2p import P2P, P2PError
from unaiverse.networking.node.tokens import TokenVerifier


class ConnectionPools:
    DEBUG = True

    def __init__(self, max_connections: int, pool_name_to_p2p_name_and_ratio: dict[str, [str, float]],
                 p2p_name_to_p2p: dict[str, P2P], public_key: str | None = None, token: str | None = None):
        """Initializes a new instance of the ConnectionPools class.

        Args:
            max_connections: The maximum total number of connections allowed across all pools.
            pool_name_to_p2p_name_and_ratio: A dictionary mapping pool names to a list containing the associated P2P network name and its connection ratio.
            p2p_name_to_p2p: A dictionary mapping P2P network names to their corresponding P2P objects.
            public_key: An optional public key for token verification.
            token: An optional initial token for authentication.

        Returns:
            None.

        """
        # Common terms: a "pool triple" is [pool_contents, max_connections_in_such_a_pool, p2p_object_of_the_pool
        self.max_con = max_connections
        self.pool_count = len(pool_name_to_p2p_name_and_ratio)
        self.pool_names = list(pool_name_to_p2p_name_and_ratio.keys())
        self.pool_ratios = [p2p_name_and_ratio[1] for p2p_name_and_ratio in pool_name_to_p2p_name_and_ratio.values()]

        # Indices involving the P2P object or its name
        self.p2p_name_to_p2p = p2p_name_to_p2p
        self.p2p_name_and_pool_name_to_pool_triple = {}
        self.p2p_to_pool_names = {}

        # Indices rooted around the pool name
        self.pool_name_to_pool_triple = {}
        self.pool_name_to_added_in_last_update = {}
        self.pool_name_to_removed_in_last_update = {}
        self.pool_name_to_peer_infos = {p: {} for p in self.pool_names}

        # Indices rooted around the peer ID
        self.peer_id_to_pool_name = {}
        self.peer_id_to_p2p = {}
        self.peer_id_to_misc = {}
        self.peer_id_to_token = {}

        # Token-related stuff, super private
        self.__token = token if token is not None else ""
        self.__token_verifier = TokenVerifier(public_key) if public_key is not None else None

        # Checking
        for p2p_name_and_ratio in pool_name_to_p2p_name_and_ratio.values():
            assert p2p_name_and_ratio[0] in self.p2p_name_to_p2p, f"Cannot find p2p named {p2p_name_and_ratio[0]} "
        assert self.max_con >= len(self.pool_names), "Too small number of max connections"
        assert sum([x for x in self.pool_ratios if x > 0]) == 1.0, "Pool ratios must sum to 1.0"

        # Preparing the pool triples
        self.pool_name_to_pool_triple = \
            {k: [set(), 0, self.p2p_name_to_p2p[pool_name_to_p2p_name_and_ratio[k][0]]] for k in self.pool_names}
        num_zero_ratio_pools = len([x for x in self.pool_ratios if x == 0])
        assert num_zero_ratio_pools <= self.max_con, "Cannot create pools given the provided max connection count"

        # Edit: to solve the teacher not engaging with more than two students.
        pools_max_sizes = {k: max(math.floor(self.pool_ratios[i] * (self.max_con - num_zero_ratio_pools)),
                                  1 if self.pool_ratios[i] >= 0. else 0)
                           for i, k in enumerate(self.pool_names)}

        # Pools_max_sizes = {k: self.max_con for k in self.pool_names}

        # Fixing sizes
        tot = 0
        for i, (k, v) in enumerate(pools_max_sizes.items()):
            assert v > 0 or self.pool_ratios[i] < 0, "Cannot create pools given the provided max connection count"

        # Edit: to solve the teacher not engaging with more than two students.
            tot += v
        assert tot <= self.max_con, \
            "Cannot create pools given the provided max connection count"
        pools_max_sizes[self.pool_names[-1]] += (self.max_con - tot)

        # Storing fixed sizes in the previously created pool triples & building additional index
        for pool_name, pool_contents_max_con_and_p2p in self.pool_name_to_pool_triple.items():
            pool_contents_max_con_and_p2p[1] = pools_max_sizes[pool_name]  # Fixing the second element of the triple

            pool, _, p2p = pool_contents_max_con_and_p2p
            p2p_name = None
            for k, v in self.p2p_name_to_p2p.items():
                if v == p2p:
                    p2p_name = k
                    break
            if p2p_name not in self.p2p_name_and_pool_name_to_pool_triple:
                self.p2p_name_and_pool_name_to_pool_triple[p2p_name] = {}
                self.p2p_to_pool_names[p2p] = []
            self.p2p_name_and_pool_name_to_pool_triple[p2p_name][pool_name] = (
                pool_contents_max_con_and_p2p)
            self.p2p_to_pool_names[p2p].append(pool_name)

    def __str__(self):
        """Returns a human-readable string representation of the connection pools' status.

        Args:
            None.

        Returns:
            A formatted string showing the number of connections and peer IDs in each pool.
        """
        max_len = max(len(s) for s in self.pool_names)
        s = f"[ConnectionPool] max_con={self.max_con}, pool_count={self.pool_count}"
        for k, (pool, max_con, _) in self.pool_name_to_pool_triple.items():
            kk = k + ","
            s += f"\n\tpool={kk.ljust(max_len + 1)}\tmax_con={max_con},\tcurrent_con={len(pool)},\tpeer_ids="
            ss = "("
            first = True
            for c in pool:
                if not first:
                    ss += ", "
                ss += c
                first = False
            ss += ")"
            s += ss
        return s

    def __getitem__(self, p2p_name):
        """Retrieves a P2P object by its name.

        Args:
            p2p_name: The name of the P2P network.

        Returns:
            The P2P object.
        """
        return self.p2p_name_to_p2p[p2p_name]

    def conn_routing_fcn(self, connected_peer_infos: list, p2p: P2P):
        """A placeholder function that must be implemented to route connected peers to the correct pool.

        Args:
            connected_peer_infos: A list of dictionaries containing information about connected peers.
            p2p: The P2P network object from which the peers were connected.

        Returns:
            A dictionary mapping pool names to a dictionary of peer IDs and their information.
        """
        raise NotImplementedError("You must implement conn_routing_fcn!")

    @staticmethod
    def __connect(p2p: P2P, addresses: list[str]):
        """Establishes a connection to a peer via a P2P network.

        Args:
            p2p: The P2P network object to use for the connection.
            addresses: A list of addresses of the peer to connect to.

        Returns:
            A tuple containing the peer ID and a boolean indicating if the connection was established through a relay.
        """
        if ConnectionPools.DEBUG:
            print(f"[DEBUG CONNECTIONS-POOL] Connecting to {addresses}")

        try:
            winning_addr_info_dict = p2p.connect_to(addresses)
            peer_id = winning_addr_info_dict.get('ID')
            connected_addr_str = winning_addr_info_dict.get('Addrs')[0]
            through_relay = '/p2p-circuit/' in connected_addr_str
            if ConnectionPools.DEBUG:
                print(f"[DEBUG CONNECTIONS-POOL] Connected to peer {peer_id} via {connected_addr_str} "
                      f"(through relay: {through_relay})")

            return peer_id, through_relay
        except P2PError:
            if ConnectionPools.DEBUG:
                print(f"[DEBUG CONNECTIONS-POOL] Connection failed!")
            return None, False

    @staticmethod
    def disconnect(p2p: P2P, peer_id: str):
        """Disconnects from a specific peer on a P2P network.

        Args:
            p2p: The P2P network object to use for disconnection.
            peer_id: The peer ID to disconnect from.

        Returns:
            True if the disconnection is successful, otherwise False.
        """
        try:
            p2p.disconnect_from(peer_id)
        except P2PError:
            return False
        return True

    def set_token(self, token: str):
        """Sets the authentication token for the connection pools.

        Args:
            token: The new token string.
        """
        self.__token = token

    def verify_token(self, token: str, peer_id: str):
        """Verifies a received token using the provided public key.

        Args:
            token: The token string to verify.
            peer_id: The peer ID associated with the token.

        Returns:
            A tuple containing the node ID and CV hash if the token is valid, otherwise None.
        """
        if self.__token_verifier is None:
            return None
        else:
            node_id, cv_hash = self.__token_verifier.verify_token(token, p2p_peer=peer_id)
            return node_id, cv_hash  # If the verification fails, this is None, None

    def connect(self, addresses: list[str], p2p_name: str):
        """Connects to a peer on a specified P2P network.

        Args:
            addresses: A list of addresses of the peer to connect to.
            p2p_name: The name of the P2P network to use.

        Returns:
            A tuple containing the peer ID of the connected peer and a boolean indicating if a relay was used.
        """
        p2p = self.p2p_name_to_p2p[p2p_name]

        # Connecting
        peer_id, through_relay = ConnectionPools.__connect(p2p, addresses)
        return peer_id, through_relay

    def add(self, peer_info: dict, pool_name: str):
        """Adds a connected peer to a specified connection pool.

        Args:
            peer_info: A dictionary containing information about the peer.
            pool_name: The name of the pool to add the peer to.

        Returns:
            True if the peer is successfully added, otherwise False.
        """
        peer_id = peer_info['id']
        pool, max_size, p2p = self.pool_name_to_pool_triple[pool_name]
        if len(pool) < max_size:

            # "hoping" peer IDs are unique, and stopping duplicate cases
            if peer_id in self.peer_id_to_pool_name and self.peer_id_to_pool_name[peer_id] != pool_name:
                return False

            self.peer_id_to_pool_name[peer_id] = pool_name
            self.peer_id_to_p2p[peer_id] = p2p

            # Setting 'misc' field (default is 0, where 0 means public)
            peer_info['misc'] = self.peer_id_to_misc.get(peer_id, 0)

            # Storing (only)
            pool.add(peer_id)
            self.pool_name_to_peer_infos[pool_name][peer_id] = peer_info
            return True
        else:
            return False

    def remove(self, peer_id: str):
        """Removes a peer from its connection pool and disconnects from it.

        Args:
            peer_id: The peer ID to remove.

        Returns:
            True if the peer is successfully removed, otherwise False.
        """
        if peer_id in self.peer_id_to_pool_name:
            pool_name = self.peer_id_to_pool_name[peer_id]
            pool, _, p2p = self.pool_name_to_pool_triple[pool_name]

            # Disconnecting
            disc = ConnectionPools.disconnect(p2p, peer_id)
            pool.remove(peer_id)
            del self.pool_name_to_peer_infos[pool_name][peer_id]
            del self.peer_id_to_pool_name[peer_id]
            del self.peer_id_to_p2p[peer_id]
            if peer_id in self.peer_id_to_misc:
                del self.peer_id_to_misc[peer_id]
            if peer_id in self.peer_id_to_token:
                del self.peer_id_to_token[peer_id]
            return disc
        else:
            return False

    def get_all_connected_peer_infos(self, pool_name: str):
        """Retrieves a list of peer information dictionaries for a given pool.

        Args:
            pool_name: The name of the pool to query.

        Returns:
            A list of dictionaries, each containing information about a peer in the pool.
        """
        return list(self.pool_name_to_peer_infos[pool_name].values())

    def get_pool_status(self):
        """Returns a dictionary showing the set of peer IDs in each pool.

        Args:
            None.

        Returns:
            A dictionary mapping pool names to the set of peer IDs in that pool.
        """
        return {k: v[0] for k, v in self.pool_name_to_pool_triple.items()}

    def get_all_connected_peer_ids(self):
        """Retrieves a list of all peer IDs currently connected across all pools.

        Args:
            None.

        Returns:
            A list of all connected peer IDs.
        """
        return list(self.peer_id_to_pool_name.keys())

    def update(self):
        """Refreshes the connection pools by checking for new and lost connections.

        Args:
            None.

        Returns:
            A tuple containing two dictionaries: one for newly added peers and one for removed peers, both keyed by
            pool name.
        """
        self.pool_name_to_added_in_last_update = {}
        self.pool_name_to_removed_in_last_update = {}

        for p2p_name, p2p in self.p2p_name_to_p2p.items():
            connected_peer_infos = p2p.get_connected_peers_info()

            if connected_peer_infos is not None:

                # Routing to the right queue / filtering
                pool_name_and_peer_ids_to_peer_info = self.conn_routing_fcn(connected_peer_infos, p2p)

                # Parsing the generated index
                for pool_name, connected_peer_ids_to_connected_peer_infos \
                        in pool_name_and_peer_ids_to_peer_info.items():
                    pool, _, pool_p2p = self.p2p_name_and_pool_name_to_pool_triple[p2p_name][pool_name]
                    connected_peer_ids = connected_peer_ids_to_connected_peer_infos.keys()
                    new_peer_ids = connected_peer_ids - pool
                    lost_peer_ids = pool - connected_peer_ids

                    # Clearing disconnected agents
                    for lost_peer_id in lost_peer_ids:
                        self.pool_name_to_removed_in_last_update.setdefault(pool_name, set()).add(lost_peer_id)

                    # Adding new agents
                    for new_peer_id in new_peer_ids:
                        peer_info = connected_peer_ids_to_connected_peer_infos[new_peer_id]
                        if not self.add(peer_info, pool_name=pool_name):
                            break
                        self.pool_name_to_added_in_last_update.setdefault(pool_name, set()).add(new_peer_id)

        return self.pool_name_to_added_in_last_update, self.pool_name_to_removed_in_last_update

    def get_messages(self, p2p_name: str, allowed_not_connected_peers: set | None = None):
        """Retrieves and verifies all messages from a specified P2P network.

        Args:
            p2p_name: The name of the P2P network to fetch messages from.
            allowed_not_connected_peers: An optional set of peer IDs to allow messages from, even if they are not in the pools.

        Returns:
            A list of verified and processed message objects.
        """
        # Pop all messages
        messages: list[Msg] = self[p2p_name].pop_messages()  # Pop all messages (list of messages - list[Msg])
        ret = []
        for m in messages:
            if (m.sender in self.peer_id_to_pool_name or  # Check if expected sender
                    (allowed_not_connected_peers is not None and m.sender in allowed_not_connected_peers)):
                try:
                    token_with_inspector_final_bit = m.piggyback
                    token = token_with_inspector_final_bit[0:-1]
                    inspector_mode = token_with_inspector_final_bit[-1]
                    node_id, _ = self.verify_token(token, m.sender)
                    if node_id is not None:

                        # Replacing piggyback with the node ID and the flag telling if it is inspector
                        m.piggyback = node_id + inspector_mode
                        ret.append(m)
                        if m.sender in self.peer_id_to_pool_name:
                            self.peer_id_to_token[m.sender] = token
                    else:
                        print("Received a message missing expected info in the token payload (discarding it)")
                except Exception as e:
                    print(f"Received a message with an invalid piggyback token! (discarding it) [{e}]")
            else:
                if ConnectionPools.DEBUG:
                    print("Received a message from a unknown sender (discarding it)")
        return ret

    def get_added_after_updating(self, pool_name: str | None = None):
        """Retrieves the peers that were added in the last update cycle.

        Args:
            pool_name: The name of a specific pool to query. If None, returns data for all pools.

        Returns:
            A set of added peer IDs for the specified pool, or a dictionary of sets for all pools.
        """
        if pool_name is not None:
            return self.pool_name_to_added_in_last_update[pool_name]
        else:
            return self.pool_name_to_added_in_last_update

    def get_removed_after_updating(self, pool_name: str | None = None):
        """Retrieves the peers that were removed in the last update cycle.

        Args:
            pool_name: The name of a specific pool to query. If None, returns data for all pools.

        Returns:
            A set of removed peer IDs for the specified pool, or a dictionary of sets for all pools.
        """
        if pool_name is not None:
            return self.pool_name_to_removed_in_last_update[pool_name]
        else:
            return self.pool_name_to_removed_in_last_update

    def get_last_token(self, peer_id):
        """Retrieves the last known token for a given peer.

        Args:
            peer_id: The peer ID to query.

        Returns:
            The token string if found, otherwise None.
        """
        return self.peer_id_to_token[peer_id] if peer_id in self.peer_id_to_token else None

    def is_connected(self, peer_id: str, pool_name: str | None = None):
        """Checks if a peer is currently connected, optionally in a specific pool.

        Args:
            peer_id: The peer ID to check.
            pool_name: An optional pool name to check within.

        Returns:
            True if the peer is connected, otherwise False.
        """
        if pool_name is None:
            return peer_id in self.peer_id_to_pool_name
        else:
            return peer_id in self.peer_id_to_pool_name and pool_name == self.peer_id_to_pool_name[peer_id]

    def get_pool_of(self, peer_id: str):
        """Gets the pool name for a given connected peer.

        Args:
            peer_id: The peer ID to query.

        Returns:
            The name of the pool the peer is in.
        """
        return self.peer_id_to_pool_name[peer_id]

    def size(self, pool_name: str | None = None):
        """Returns the number of connections in a specific pool or the total number across all pools.

        Args:
            pool_name: An optional pool name to get the size of. If None, returns the total size.

        Returns:
            The size of the pool or the total number of connections.
        """
        if pool_name is not None:
            return len(self.pool_name_to_pool_triple[pool_name])
        else:
            c = 0
            for v in self.pool_name_to_pool_triple.values():
                c += len(v)
            return c

    def send(self, peer_id: str, channel_trail: str | None,
             content_type: str, content: bytes | dict | None = None, p2p: P2P | None = None):
        """Sends a direct message to a specific peer.

        Args:
            peer_id: The peer ID to send the message to.
            channel_trail: An optional string to append to the channel name.
            content_type: The type of content in the message.
            content: The message content.
            p2p: An optional P2P object to use for sending. If None, it is derived from the peer_id.

        Returns:
            True if the message is sent successfully, otherwise False.
        """
        # Getting the right p2p object
        if p2p is None:
            p2p = self.peer_id_to_p2p[peer_id] if peer_id in self.peer_id_to_p2p else None
            if p2p is None:
                if ConnectionPools.DEBUG:
                    print("[DEBUG CONNECTIONS-POOL] P2P non found for peer id: " + str(peer_id))
                return False

        # Defining channel
        if channel_trail is not None and len(channel_trail) > 0:
            channel = f"{p2p.peer_id}::dm:{peer_id}-{content_type}~{channel_trail}"
        else:
            channel = f"{p2p.peer_id}::dm:{peer_id}-{content_type}"

        # Adding sender info here
        msg = Msg(sender=p2p.peer_id,
                  content_type=content_type,
                  content=content,
                  channel=channel,
                  piggyback=self.__token + "0")  # Adding inspector-mode bit (dummy bit here)
        if ConnectionPools.DEBUG:
            print("[DEBUG CONNECTIONS-POOL] Sending message: " + str(msg))

        # Sending direct message
        try:
            p2p.send_message_to_peer(channel, msg)

            # If the line above executes without raising an error, it was successful.
            return True
        except P2PError as e:

            # If send_message_to_peer fails, it will raise a P2PError. We catch it here.
            if ConnectionPools.DEBUG:
                print("[DEBUG CONNECTIONS-POOL] Sending error is: " + str(e))
            return False

    def subscribe(self, peer_id: str, channel: str, default_p2p_name: str | None = None):
        """Subscribes to a topic/channel on a P2P network.

        Args:
            peer_id: The peer ID associated with the topic/channel.
            channel: The name of the channel to subscribe to.
            default_p2p_name: An optional P2P network name to use if the peer's network is unknown.

        Returns:
            True if the subscription is successful, otherwise False.
        """

        # Getting the right p2p object
        p2p = None
        for _p2p in self.p2p_to_pool_names.keys():
            if _p2p.peer_id == peer_id:
                p2p = _p2p
                break
        if p2p is None and peer_id in self.peer_id_to_p2p:
            p2p = self.peer_id_to_p2p[peer_id]
        if p2p is None:
            if default_p2p_name is not None:
                p2p = self.p2p_name_to_p2p[default_p2p_name]
            else:
                return False

        try:
            p2p.subscribe_to_topic(channel)
        except (P2PError, ValueError) as e:
            return False
        return True

    def unsubscribe(self, peer_id: str, channel: str):
        """Unsubscribes from a topic/channel on a P2P network.

        Args:
            peer_id: The peer ID associated with the topic/channel.
            channel: The name of the channel to unsubscribe from.

        Returns:
            True if the unsubscription is successful, otherwise False.
        """

        # Getting the right p2p object
        p2p = None
        for _p2p in self.p2p_to_pool_names.keys():
            if _p2p.peer_id == peer_id:
                p2p = _p2p
                break
        if p2p is None and peer_id in self.peer_id_to_p2p:
            p2p = self.peer_id_to_p2p[peer_id]
        if p2p is None:
            return False

        try:
            p2p.unsubscribe_from_topic(channel)
        except (P2PError, ValueError):
            return False
        return True

    def publish(self, peer_id: str, channel: str,
                content_type: str, content: bytes | dict | tuple | None = None):
        """Publishes a message to a topic/channel on a P2P network.

        Args:
            peer_id: The peer ID associated with the topic/channel.
            channel: The name of the channel to publish to.
            content_type: The type of content in the message.
            content: The message content.

        Returns:
            True if the message is published successfully, otherwise False.
        """

        # Getting the right p2p object
        p2p = None
        for _p2p in self.p2p_to_pool_names.keys():
            if _p2p.peer_id == peer_id:
                p2p = _p2p
                break
        if p2p is None:
            p2p = self.peer_id_to_p2p[peer_id]
        if p2p is None:
            return False

        # Adding sender info here
        msg = Msg(sender=p2p.peer_id,
                  content_type=content_type,
                  content=content,
                  channel=channel,
                  piggyback=self.__token + "0")  # Adding inspector-mode bit (dummy bit here)
        if ConnectionPools.DEBUG:
            print("[DEBUG CONNECTIONS-POOL] Sending (publish) message: " + str(msg))

        # Sending message via GossipSub
        try:
            p2p.broadcast_message(channel, msg)

            # If the line above executes without raising an error, it was successful.
            return True
        except P2PError as e:

            # If send_message_to_peer fails, it will raise a P2PError. We catch it here.
            return False


class NodeConn(ConnectionPools):

    # Basic name
    __ALL_UNIVERSE = "all_universe"
    __WORLD_AGENTS_ONLY = "world_agents"
    __WORLD_NODE_ONLY = "world_node"
    __WORLD_MASTERS_ONLY = "world_masters"

    # Suffixes
    __PUBLIC_NET = "_public"
    __PRIVATE_NET = "_private"

    # Prefixes
    __INBOUND = "in_"
    __OUTBOUND = "out_"

    # P2p names
    P2P_PUBLIC = "p2p_public"
    P2P_WORLD = "p2p_world"

    # All pools (prefix + basic name + suffix)
    IN_PUBLIC = __INBOUND + __ALL_UNIVERSE + __PUBLIC_NET
    OUT_PUBLIC = __OUTBOUND + __ALL_UNIVERSE + __PUBLIC_NET
    IN_WORLD_AGENTS = __INBOUND + __WORLD_AGENTS_ONLY + __PRIVATE_NET
    OUT_WORLD_AGENTS = __OUTBOUND + __WORLD_AGENTS_ONLY + __PRIVATE_NET
    IN_WORLD_NODE = __INBOUND + __WORLD_NODE_ONLY + __PRIVATE_NET
    OUT_WORLD_NODE = __OUTBOUND + __WORLD_NODE_ONLY + __PRIVATE_NET
    IN_WORLD_MASTERS = __INBOUND + __WORLD_MASTERS_ONLY + __PRIVATE_NET
    OUT_WORLD_MASTERS = __OUTBOUND + __WORLD_MASTERS_ONLY + __PRIVATE_NET

    # Aggregated pools
    PUBLIC = {IN_PUBLIC, OUT_PUBLIC}
    WORLD_NODE = {IN_WORLD_NODE, OUT_WORLD_NODE}
    WORLD_AGENTS = {IN_WORLD_AGENTS, OUT_WORLD_AGENTS}
    WORLD_MASTERS = {IN_WORLD_MASTERS, OUT_WORLD_MASTERS}
    WORLD = WORLD_NODE | WORLD_AGENTS | WORLD_MASTERS
    ALL = PUBLIC | WORLD
    OUTGOING = {OUT_PUBLIC, OUT_WORLD_NODE, OUT_WORLD_AGENTS, OUT_WORLD_MASTERS}
    INCOMING = {IN_PUBLIC, IN_WORLD_NODE, IN_WORLD_AGENTS, IN_WORLD_MASTERS}

    def __init__(self, max_connections: int, p2p_u: P2P, p2p_w: P2P,
                 is_world_node: bool, public_key: str, token: str):
        """Initializes a new instance of the NodeConn class.

        Args:
            max_connections: The total number of connections the node can handle.
            p2p_u: The P2P object for the public network.
            p2p_w: The P2P object for the world/private network.
            is_world_node: A boolean flag indicating if this node is a world node.
            public_key: The public key for token verification.
            token: The node's authentication token.
        """
        super().__init__(max_connections=max_connections,
                         p2p_name_to_p2p={
                             NodeConn.P2P_PUBLIC: p2p_u,
                             NodeConn.P2P_WORLD: p2p_w,
                         },
                         pool_name_to_p2p_name_and_ratio={
                             NodeConn.IN_PUBLIC: [NodeConn.P2P_PUBLIC, 0.25 / 2. if not is_world_node else 0.25 / 2.],
                             NodeConn.OUT_PUBLIC: [NodeConn.P2P_PUBLIC, 0.25 / 2. if not is_world_node else 0.25 / 2.],
                             NodeConn.IN_WORLD_AGENTS: [NodeConn.P2P_WORLD, .75 / 2 if not is_world_node else 0.5 / 2],
                             NodeConn.OUT_WORLD_AGENTS: [NodeConn.P2P_WORLD, .75 / 2 if not is_world_node else 0.5 / 2],
                             NodeConn.IN_WORLD_NODE: [NodeConn.P2P_WORLD, 0. if not is_world_node else -1.],
                             NodeConn.OUT_WORLD_NODE: [NodeConn.P2P_WORLD, 0. if not is_world_node else -1],
                             NodeConn.IN_WORLD_MASTERS: [NodeConn.P2P_WORLD, 0. if not is_world_node else 0.25 / 2.],
                             NodeConn.OUT_WORLD_MASTERS: [NodeConn.P2P_WORLD, 0. if not is_world_node else 0.25 / 2.]
                         },
                         public_key=public_key, token=token)

        # Just for convenience
        self.p2p_public = p2p_u
        self.p2p_world = p2p_w

        # These are the list of all the possible agents that might try to connect when we are in world
        self.world_agents_list = set()
        self.world_masters_list = set()
        self.world_agents_and_world_masters_list = set()
        self.world_node_peer_id = None
        self.inspector_peer_id = None
        self.role_to_peer_ids = {}
        self.peer_id_to_addrs = {}

        # Rendezvous
        self.rendezvous_tag = -1

    def reset_rendezvous_tag(self):
        """Resets the rendezvous tag to its initial state."""
        self.rendezvous_tag = -1

    def conn_routing_fcn(self, connected_peer_infos: list, p2p: P2P):
        """Routes connected peers to the correct connection pool based on their network and role.

        Args:
            connected_peer_infos: A list of dictionaries with information about connected peers.
            p2p: The P2P network object where the connections were found.

        Returns:
            A dictionary mapping pool names to a dictionary of peer IDs and their information.
        """
        pool_name_and_peer_id_to_peer_info = {k: {} for k in self.p2p_to_pool_names[p2p]}
        public = p2p == self.p2p_public

        for c in connected_peer_infos:
            inbound = c['direction'] == "incoming"
            outbound = c['direction'] == "outgoing"
            peer_id = c['id']  # Other fields are: c['addrs'], c['connected_at']

            if public:
                if inbound:
                    pool_name_and_peer_id_to_peer_info[NodeConn.IN_PUBLIC][peer_id] = c
                elif outbound:
                    pool_name_and_peer_id_to_peer_info[NodeConn.OUT_PUBLIC][peer_id] = c
                else:
                    raise ValueError(f"Connection direction is undefined: {c['direction']}")
            else:
                is_world_agent = peer_id in self.world_agents_list
                is_world_master = peer_id in self.world_masters_list
                is_world_node = self.world_node_peer_id is not None and peer_id == self.world_node_peer_id
                is_inspector = self.inspector_peer_id is not None and peer_id == self.inspector_peer_id
                if not is_world_node and not is_world_master and not is_world_agent and not is_inspector:
                    if ConnectionPools.DEBUG:
                        print("[DEBUG CONNECTIONS-POOL] World agents list:  " + str(self.world_agents_list))
                        print("[DEBUG CONNECTIONS-POOL] World masters list: " + str(self.world_masters_list))
                        print("[DEBUG CONNECTIONS-POOL] World node peer id: " + str(self.world_node_peer_id))
                        print("[DEBUG CONNECTIONS-POOL] Inspector peer id: " + str(self.inspector_peer_id))
                        print(f"[DEBUG CONNECTIONS-POOL] Unable to determine the peer type for {peer_id}: "
                              f"cannot say if world agent, master, world node, inspector (disconnecting it)")
                    ConnectionPools.disconnect(p2p, peer_id)
                    continue

                if inbound:
                    pool_name_and_peer_id_to_peer_info[NodeConn.IN_WORLD_AGENTS if is_world_agent else (
                            NodeConn.IN_WORLD_NODE if is_world_node else
                            NodeConn.IN_WORLD_MASTERS)][peer_id] = c
                elif outbound:
                    pool_name_and_peer_id_to_peer_info[NodeConn.OUT_WORLD_AGENTS if is_world_agent else (
                            NodeConn.OUT_WORLD_NODE if is_world_node else
                            NodeConn.OUT_WORLD_MASTERS)][peer_id] = c
                else:
                    raise ValueError(f"Connection direction is undefined: {c}")

        return pool_name_and_peer_id_to_peer_info

    def set_world(self, world_peer_id: str | None):
        """Sets the peer ID of the world node.

        Args:
            world_peer_id: The peer ID of the world node, or None to clear it.
        """
        self.world_node_peer_id = world_peer_id

    def set_inspector(self, inspector_peer_id: str | None):
        """Sets the peer ID of the inspector.

        Args:
            inspector_peer_id: The peer ID of the inspector node.
        """
        self.inspector_peer_id = inspector_peer_id

    def get_world_peer_id(self):
        """Returns the peer ID of the world node.

        Args:
            None.

        Returns:
            The world node's peer ID.
        """
        return self.world_node_peer_id

    def set_addresses_in_peer_info(self, peer_id, addresses):
        """Updates the list of addresses for a given peer.

        Args:
            peer_id: The peer ID to update.
            addresses: A new list of addresses for the peer.
        """
        if self.in_connection_queues(peer_id):
            addrs = self.pool_name_to_peer_infos[self.get_pool_of(peer_id)][peer_id]['addrs']
            addrs.clear()  # Warning: do not allocate a new list, keep the current one (it is referenced by others)
            for _addrs in addresses:
                addrs.append(_addrs)

    def set_role(self, peer_id, new_role: int):
        """Updates the role of a peer and its associated role-based lists.

        Args:
            peer_id: The peer ID to update.
            new_role: The new role for the peer.
        """
        cur_role = self.get_role(peer_id)

        # Updating
        self.peer_id_to_misc[peer_id] = new_role

        if self.in_connection_queues(peer_id):
            self.pool_name_to_peer_infos[self.get_pool_of(peer_id)][peer_id]['misc'] = new_role

        # Updating
        if cur_role in self.role_to_peer_ids:
            if peer_id in self.role_to_peer_ids[cur_role]:
                self.role_to_peer_ids[cur_role].remove(peer_id)
            if len(self.role_to_peer_ids[cur_role]) == 0:
                del self.role_to_peer_ids[cur_role]
        if new_role not in self.role_to_peer_ids:
            self.role_to_peer_ids[new_role] = set()
        self.role_to_peer_ids[new_role].add(peer_id)

    def set_world_agents_list(self, world_agents_list_peer_infos: list[dict] | None):
        """Sets the list of all world agents based on a provided list of peer information.

        Args:
            world_agents_list_peer_infos: A list of dictionaries containing peer information for world agents.
        """

        # Clearing previous information
        to_remove = []
        for peer_id, misc in self.peer_id_to_misc.items():
            if misc & 1 == 1 and misc & 2 == 0:
                to_remove.append((peer_id, misc))

        for peer_id, misc in to_remove:
            del self.peer_id_to_misc[peer_id]
            if peer_id in self.peer_id_to_addrs:
                del self.peer_id_to_addrs[peer_id]
            self.role_to_peer_ids[misc].discard(peer_id)

        # Setting new information
        if world_agents_list_peer_infos is not None and len(world_agents_list_peer_infos) > 0:
            self.world_agents_list = {x['id'] for x in world_agents_list_peer_infos}
            for x in world_agents_list_peer_infos:
                self.peer_id_to_addrs[x['id']] = x['addrs']
                self.set_role(x['id'], x['misc'])
        else:
            self.world_agents_list = set()

        self.world_agents_and_world_masters_list = self.world_agents_list | self.world_masters_list

    def set_world_masters_list(self, world_masters_list_peer_infos: list[dict] | None):
        """Sets the list of all world masters based on a provided list of peer information.

        Args:
            world_masters_list_peer_infos: A list of dictionaries containing peer information for world masters.
        """

        # Clearing previous information
        to_remove = []
        for peer_id, misc in self.peer_id_to_misc.items():
            if misc & 1 == 1 and misc & 2 == 2:
                to_remove.append((peer_id, misc))

        for peer_id, misc in to_remove:
            del self.peer_id_to_misc[peer_id]
            if peer_id in self.peer_id_to_addrs:
                del self.peer_id_to_addrs[peer_id]
            self.role_to_peer_ids[misc].discard(peer_id)

        # Setting new information
        if world_masters_list_peer_infos is not None and len(world_masters_list_peer_infos) > 0:
            self.world_masters_list = {x['id'] for x in world_masters_list_peer_infos}
            for x in world_masters_list_peer_infos:
                self.peer_id_to_addrs[x['id']] = x['addrs']
                self.set_role(x['id'], x['misc'])
        else:
            self.world_masters_list = set()

        self.world_agents_and_world_masters_list = self.world_agents_list | self.world_masters_list

    def add_to_world_agents_list(self, peer_id: str, addrs: list[str], role: int = -1):
        """Adds a new world agent to the list.

        Args:
            peer_id: The peer ID of the new agent.
            addrs: A list of addresses for the new agent.
            role: The role assigned to the agent.
        """
        self.world_agents_list.add(peer_id)

        # This assumes that the WORLD MASTER/AGENT BIT is the first one
        assert role & 1 == 1, "Expecting the first bit of the role to be 1 for world agents"
        assert role & 2 == 0, "Expecting the second bit of the role to be 0 for world agents"
        self.peer_id_to_addrs[peer_id] = addrs
        self.set_role(peer_id, role)
        self.world_agents_and_world_masters_list = self.world_agents_list | self.world_masters_list

    def add_to_world_masters_list(self, peer_id: str, addrs: list[str], role: int = -1):
        """Adds a new world master to the list.

        Args:
            peer_id: The peer ID of the new master.
            addrs: A list of addresses for the new master.
            role: The role assigned to the master.
        """
        self.world_masters_list.add(peer_id)

        # This assumes that the WORLD MASTER/AGENT BIT is the first one
        assert role & 1 == 1, "Expecting the first bit of the role to be 1 for world masters"
        assert role & 2 == 2, "Expecting the second bit of the role to be 1 for world masters"
        self.peer_id_to_addrs[peer_id] = addrs
        self.set_role(peer_id, role)
        self.world_agents_and_world_masters_list = self.world_agents_list | self.world_masters_list

    def get_added_after_updating(self, pool_names: list[str] | None = None):
        """Retrieves the set of peers added after the last update cycle for specified pools.

        Args:
            pool_names: A list of pool names to check. If None, checks all pools.

        Returns:
            A dictionary mapping pool names to sets of added peer IDs, or a single set if only one pool is specified.
        """
        if pool_names is not None:
            ret = {}
            for p in pool_names:
                ret[p] = super().get_added_after_updating(p)
            return ret
        else:
            return super().get_added_after_updating()

    def get_removed_after_updating(self, pool_names: list[str] | None = None):
        """Retrieves the set of peers removed after the last update cycle for specified pools.

        Args:
            pool_names: A list of pool names to check. If None, checks all pools.

        Returns:
            A dictionary mapping pool names to sets of removed peer IDs, or a single set if only one pool is specified.
        """
        if pool_names is not None:
            ret = {}
            for p in pool_names:
                ret[p] = super().get_removed_after_updating(p)
            return ret
        else:
            return super().get_removed_after_updating()

    def size(self, pool_names: list[str] | None = None):
        """Returns the total number of connections across all specified pools.

        Args:
            pool_names: A list of pool names to sum the size of. If None, returns the total size of all pools.

        Returns:
            The total number of connections.
        """
        if pool_names is not None:
            return super().size()
        else:
            c = 0
            for p in self.pool_names:
                c += super().size(p)
            return c

    def is_connected(self, peer_id: str, pool_names: list[str] | None = None):
        """Checks if a peer is connected in any of the specified pools.

        Args:
            peer_id: The peer ID to check.
            pool_names: A list of pool names to search within. If None, searches all pools.

        Returns:
            True if the peer is found in any of the pools, otherwise False.
        """
        if pool_names is None:
            return super().is_connected(peer_id)
        else:
            for p in pool_names:
                if super().is_connected(peer_id, p):
                    return True
            return False

    def is_public(self, peer_id):
        """Checks if a peer is connected via the public network.

        Args:
            peer_id: The peer ID to check.

        Returns:
            True if the peer is in a public pool, otherwise False.
        """
        pool_name = self.get_pool_of(peer_id)
        return pool_name in NodeConn.PUBLIC

    def is_world_master(self, peer_id):
        """Checks if a peer is a world master.

        Args:
            peer_id: The peer ID to check.

        Returns:
            True if the peer is in a world master pool, otherwise False.
        """
        pool_name = self.get_pool_of(peer_id)
        return pool_name in NodeConn.WORLD_MASTERS

    def is_world_node(self, peer_id):
        """Checks if a peer is the world node.

        Args:
            peer_id: The peer ID to check.

        Returns:
            True if the peer is in a world node pool, otherwise False.
        """
        pool_name = self.get_pool_of(peer_id)
        return pool_name in NodeConn.WORLD_NODE

    def is_in_world(self, peer_id):
        """Checks if a peer is connected to the world network.

        Args:
            peer_id: The peer ID to check.

        Returns:
            True if the peer is in any world pool, otherwise False.
        """
        pool_name = self.get_pool_of(peer_id)
        return pool_name in NodeConn.WORLD

    def get_role(self, peer_id):
        """Retrieves the role of a given peer.

        Args:
            peer_id: The peer ID to query.

        Returns:
            The integer role of the peer.
        """
        role = self.peer_id_to_misc.get(peer_id, 0)  # 0 means public
        assert role >= 0, "Expecting role to be >= 0"
        assert role & 1 != 0 or role == 0, "Expecting public role to be zero (all-zero-bits)"
        return role

    def get_addrs(self, peer_id):
        """Retrieves the list of addresses for a given peer.

        Args:
            peer_id: The peer ID to query.

        Returns:
            A list of addresses for the peer.
        """
        return self.peer_id_to_addrs.get(peer_id)

    def in_connection_queues(self, peer_id):
        """Checks if a peer ID exists in any connection pool.

        Args:
            peer_id: The peer ID to check.

        Returns:
            True if the peer is found in any pool, otherwise False.
        """
        return peer_id in self.peer_id_to_pool_name

    def find_addrs_by_role(self, role, return_peer_ids_too: bool = False):
        """Finds all addresses of peers with a specific role.

        Args:
            role: The integer role to search for.
            return_peer_ids_too: A boolean to also return the peer IDs.

        Returns:
            A list of lists of addresses, and optionally a list of peer IDs.
        """
        if role in self.role_to_peer_ids:
            peer_ids = self.role_to_peer_ids[role]
        else:
            if not return_peer_ids_too:
                return []
            else:
                return [], []
        ret_addrs = []
        ret_peer_ids = []
        for peer_id in peer_ids:
            addrs = self.get_addrs(peer_id)
            if addrs is not None:
                ret_addrs.append(addrs)
                ret_peer_ids.append(peer_id)
        if not return_peer_ids_too:
            return ret_addrs
        else:
            return ret_addrs, ret_peer_ids

    def count_by_role(self, role: int):
        """Counts the number of peers with a specific role.

        Args:
            role: The integer role to count.

        Returns:
            The number of peers with that role.
        """
        if role in self.role_to_peer_ids:
            return len(self.role_to_peer_ids[role])
        else:
            return 0

    def get_all_connected_peer_infos(self, pool_names: list[str] | set[str]):
        """Retrieves a list of all peer info dictionaries for the specified pools.

        Args:
            pool_names: A list or set of pool names to query.

        Returns:
            A list of dictionaries containing peer information.
        """
        ret = []
        for p in pool_names:
            ret += super().get_all_connected_peer_infos(p)
        return ret

    def set_world_agents_and_world_masters_lists_from_rendezvous(self):
        """Updates the lists of world agents and masters using data from the rendezvous topic."""
        rendezvous_state = self.p2p_world.get_rendezvous_peers_info()

        if rendezvous_state is not None:
            tag = rendezvous_state.get('update_count', -1)

            if tag > self.rendezvous_tag:
                self.rendezvous_tag = tag
                rendezvous_peer_infos = rendezvous_state.get('peers', [])

                world_agents_peer_infos = []
                world_masters_peer_infos = []

                if ConnectionPools.DEBUG:
                    print(f"[DEBUG CONNECTIONS-POOL] Rendezvous peer infos (tag: {tag}, peers: "
                          f"{len(rendezvous_peer_infos)} peers)")

                for c in rendezvous_peer_infos:
                    if c['addrs'] is None or len(c['addrs']) == 0:
                        print(f"[DEBUG CONNECTIONS-POOL] Skipping a peer with missing addrs: {c}")
                        continue
                    if (c['misc'] & 1) == 1 and (c['misc'] & 2) == 0:
                        world_agents_peer_infos.append(c)
                    elif (c['misc'] & 1) == 1 and (c['misc'] & 2) == 2:
                        world_masters_peer_infos.append(c)
                    else:
                        raise ValueError("Unexpected value of the 'misc' field: " + str(c))

                # Updating lists
                self.set_world_agents_list(world_agents_peer_infos)
                self.set_world_masters_list(world_masters_peer_infos)

    def get_cv_hash_from_last_token(self, peer_id):
        """Retrieves the CV hash from the last token received from a peer.

        Args:
            peer_id: The peer ID to query.

        Returns:
            The CV hash string, or None if not found.
        """
        token = self.get_last_token(peer_id)
        if token is not None:
            _, cv_hash = self.verify_token(token, peer_id)
            return cv_hash
        else:
            return None

    def remove(self, peer_id: str):
        """Removes a peer and its associated information from all lists and pools.

        Args:
            peer_id: The peer ID to remove.
        """
        super().remove(peer_id)
        if peer_id in self.peer_id_to_addrs:
            del self.peer_id_to_addrs[peer_id]

    def remove_all_world_agents(self):
        """Removes all connected world agents from the pools and role lists."""
        peer_infos = self.get_all_connected_peer_infos(NodeConn.WORLD)
        for c in peer_infos:
            peer_id = c['id']
            self.remove(peer_id)
            for role, peer_ids in self.role_to_peer_ids.items():
                if role & 1 == NodeConn.WORLD:
                    peer_ids.remove(peer_id)

    def subscribe(self, peer_id: str, channel: str, default_p2p_name: str | None = None):
        """Subscribes to a channel, defaulting to the world P2P network if a network is not specified.

        Args:
            peer_id: The peer ID associated with the channel.
            channel: The channel to subscribe to.
            default_p2p_name: An optional P2P name to use for the subscription.

        Returns:
            True if successful, False otherwise.
        """
        return super().subscribe(peer_id, channel,
                                 default_p2p_name=NodeConn.P2P_WORLD if default_p2p_name is None else default_p2p_name)

    def get_messages(self, p2p_name: str, allowed_not_connected_peers: set | None = None):
        """Retrieves messages, allowing for messages from known world agents and masters even if not in a connection pool.

        Args:
            p2p_name: The name of the P2P network to get messages from.
            allowed_not_connected_peers: This parameter is ignored in this implementation.

        Returns:
            A list of verified and processed message objects.
        """
        assert allowed_not_connected_peers is None, "This param (allowed_not_connected_peers is ignored in NodeConn"
        return super().get_messages(p2p_name, allowed_not_connected_peers=self.world_agents_and_world_masters_list)
